from dataclasses import dataclass
from typing import Any

import torch.nn as nn

from ....utils.interfaces import UdaoEmbedItemShape
from ...embedders.base_embedder import BaseEmbedder


class DummyEmbedder(BaseEmbedder):
    @dataclass
    class Params(BaseEmbedder.Params):
        input_size: int

    def __init__(self, params: Params) -> None:
        super().__init__(net_params=params)
        self.input_size = params.input_size
        self.fc_layer = nn.Linear(self.input_size, self.embedding_size)

    @classmethod
    def from_iterator_shape(
        cls,
        iterator_shape: UdaoEmbedItemShape,
        **kwargs: Any,
    ) -> "DummyEmbedder":
        return DummyEmbedder(
            params=DummyEmbedder.Params(
                input_size=iterator_shape.embedding_input_shape, **kwargs
            )
        )

    def forward(self, input: Any) -> Any:
        return input



from typing import cast

import dgl
import pytest
import torch as th

from ...embedders.graph_transformer import GraphTransformer
from ...embedders.layers.multi_head_attention import AttentionLayerName
from ...utils.utils import set_deterministic_torch
from .conftest import generate_dgl_graph


def test_graph_transformer_initialization() -> None:
    for attention_layer in ["QF", "RAAL", "GTN"]:
        net_params = GraphTransformer.Params(
            input_size=3,
            output_size=6,
            op_groups=["type", "cbo"],
            type_embedding_dim=3,
            embedding_normalizer="BN",
            n_op_types=4,
            pos_encoding_dim=5,
            n_layers=3,
            n_heads=2,
            hidden_dim=6,
            readout="sum",
            attention_layer_name=cast(AttentionLayerName, attention_layer),
            dropout=0.1,
            max_dist=3 if attention_layer == "QF" else None,
            non_siblings_map={0: {0: [0, 1, 2], 1: [1, 2, 3]}}
            if attention_layer == "RAAL"
            else None,
        )

        model = GraphTransformer(net_params)
        assert isinstance(model, GraphTransformer)


def test_graph_transformer_initialization_raises_error() -> None:
    with pytest.raises(ValueError):
        net_params = GraphTransformer.Params(
            input_size=3,
            output_size=3,
            op_groups=["type", "cbo"],
            type_embedding_dim=3,
            embedding_normalizer="BN",
            n_op_types=4,
            pos_encoding_dim=5,
            n_layers=3,
            n_heads=2,
            hidden_dim=6,
            readout="sum",
            dropout=0.1,
        )
        GraphTransformer(net_params)


def test_graph_transformer_forward() -> None:
    set_deterministic_torch(0)
    net_params = GraphTransformer.Params(
        input_size=3,
        output_size=4,
        op_groups=["type", "cbo"],
        type_embedding_dim=5,
        embedding_normalizer="BN",
        n_op_types=5,
        pos_encoding_dim=5,
        n_layers=3,
        n_heads=2,
        hidden_dim=4,
        readout="sum",
        attention_layer_name="GTN",
        dropout=0.1,
    )

    transformer = GraphTransformer(net_params)
    features = {
        "op_gid": {"type": "int", "size": 1},
        "cbo": {"type": "float", "size": 3},
    }

    g1 = generate_dgl_graph(10, 9, features)
    g2 = generate_dgl_graph(
        5,
        4,
        features,
    )

    g_batch = dgl.batch([g1, g2])
    h_lap_pos_enc = th.randn(10 + 5, 5)
    g_batch.ndata["pos_enc"] = h_lap_pos_enc
    transformer.eval()
    output = transformer.forward(g_batch)
    expected = th.tensor(
        [
            [4.184283, -2.093954, 11.841289, 16.731495],
            [1.523479, -0.436974, 7.724406, 4.564045],
        ]
    )
    assert output.shape == (2, net_params.output_size)
    assert th.allclose(output, expected, atol=1e-5)



import dgl
import pytest
import torch as th

from ...embedders.graph_averager import GraphAverager
from ...utils.utils import set_deterministic_torch
from .conftest import generate_dgl_graph


@pytest.fixture
def params_fixture() -> GraphAverager.Params:
    return GraphAverager.Params(
        input_size=2,
        output_size=4,
        op_groups=["type", "cbo"],
        type_embedding_dim=5,
        embedding_normalizer=None,
        n_op_types=3,
    )


class TestAverager:
    def test_forward_shape(self, params_fixture: GraphAverager.Params) -> None:
        set_deterministic_torch(0)
        averager = GraphAverager(params_fixture)
        features_dict = {
            "op_gid": {"size": 1, "type": "int"},
            "cbo": {"size": 2, "type": "float"},
        }
        g1 = generate_dgl_graph(3, 2, features_dict)
        g2 = generate_dgl_graph(5, 4, features_dict)
        g_batch = dgl.batch([g1, g2])
        embedding = averager.forward(g_batch)

        expected = th.tensor(
            [
                [0.399936, 0.393089, 0.000000, 0.469329],
                [0.357530, 0.572405, 0.000000, 0.614257],
            ]
        )
        assert th.isclose(embedding, expected, rtol=1e-5).all()
        assert embedding.shape == (2, 4)



class NoSolutionError(ValueError):
    "Raised when no solution is found for an MOO problem"
    ...


class UncompliantSolutionError(ValueError):
    "Raised when the candidate is not compliant with the optimization algorithm"
    ...



from typing import Dict, Optional

import numpy as np
import pytest
import torch as th

from ....utils.logging import logger
from ...concepts import Constraint, Objective
from ...concepts.problem import MOProblem
from ...moo.weighted_sum import WeightedSum
from ...soo.grid_search_solver import GridSearchSolver
from ...soo.mogd import MOGD
from ...soo.random_sampler_solver import RandomSamplerSolver
from ...soo.so_solver import SOSolver
from ...utils.exceptions import NoSolutionError
from ...utils.moo_utils import even_weights


class TestWeightedSum:
    @pytest.mark.parametrize(
        "inner_solver",
        [
            GridSearchSolver(GridSearchSolver.Params(n_grids_per_var=[2, 7])),
            RandomSamplerSolver(RandomSamplerSolver.Params(n_samples_per_param=30)),
        ],
    )
    def test_solve_without_input_parameters(
        self, inner_solver: SOSolver, simple_problem: MOProblem
    ) -> None:
        """solve a dummy minimization problem with 2 objectives and 1 constraint"""
        ws_pairs = np.array([[0.3, 0.7], [0.6, 0.4]])
        simple_problem.input_parameters = None

        ws_algo = WeightedSum(
            WeightedSum.Params(
                so_solver=inner_solver,
                ws_pairs=ws_pairs,
            )
        )
        po_objs, po_vars = ws_algo.solve(problem=simple_problem, seed=0)
        np.testing.assert_array_almost_equal(po_objs, np.array([[0, 0.2]]))
        np.testing.assert_equal(po_vars, np.array({"v1": 0, "v2": 2}))

    @pytest.mark.parametrize(
        "inner_solver",
        [
            GridSearchSolver(GridSearchSolver.Params(n_grids_per_var=[2, 7])),
            RandomSamplerSolver(RandomSamplerSolver.Params(n_samples_per_param=30)),
        ],
    )
    def test_solve_with_input_parameters(
        self, inner_solver: SOSolver, simple_problem: MOProblem
    ) -> None:
        """solve a dummy minimization problem with 2 objectives and 1 constraint"""
        ws_pairs = np.array([[0.3, 0.7], [0.6, 0.4]])

        ws_algo = WeightedSum(
            WeightedSum.Params(
                so_solver=inner_solver,
                ws_pairs=ws_pairs,
            )
        )
        po_objs, po_vars = ws_algo.solve(problem=simple_problem, seed=0)

        np.testing.assert_almost_equal(po_objs, np.array([[1, 1.3]]))
        np.testing.assert_equal(po_vars, np.array([{"v1": 0, "v2": 3}]))

    @pytest.mark.parametrize(
        "inner_solver",
        [
            GridSearchSolver(GridSearchSolver.Params(n_grids_per_var=[2, 7])),
            RandomSamplerSolver(RandomSamplerSolver.Params(n_samples_per_param=1000)),
        ],
    )
    def test_solver_with_two_obj_problem(
        self, inner_solver: SOSolver, two_obj_problem: MOProblem
    ) -> None:
        ws_pairs = np.array(
            [
                [0.3, 0.7],
                [0.6, 0.4],
                [0.1, 0.9],
                [0.2, 0.8],
                [0.4, 0.6],
                [0.5, 0.5],
            ]
        )

        ws_algo = WeightedSum(
            WeightedSum.Params(
                so_solver=inner_solver,
                ws_pairs=ws_pairs,
            )
        )
        po_objs, po_vars = ws_algo.solve(problem=two_obj_problem, seed=0)

        np.testing.assert_almost_equal(po_objs, np.array([[0, 0]]), decimal=5)
        np.testing.assert_almost_equal(po_vars[0]["v1"], 0.0, decimal=3)
        assert po_vars[0]["v2"] == 1.0

    @pytest.mark.parametrize(
        "strict_rounding",
        [
            True,
            False,
        ],
    )
    def test_solver_with_two_obj_problem_mogd(
        self, strict_rounding: bool, two_obj_problem: MOProblem
    ) -> None:
        inner_solver = MOGD(
            MOGD.Params(
                learning_rate=0.1,
                max_iters=100,
                patience=20,
                multistart=2,
                batch_size=10,
                strict_rounding=strict_rounding,
            )
        )

        ws_pairs = even_weights(0.1, 2)
        logger.debug(f"ws_pairs: {ws_pairs}")
        ws_algo = WeightedSum(
            WeightedSum.Params(
                so_solver=inner_solver,
                ws_pairs=ws_pairs,
                normalize=False,
            )
        )
        po_objs, po_vars = ws_algo.solve(problem=two_obj_problem, seed=0)

        np.testing.assert_almost_equal(po_objs, np.array([[0, 0]]))
        np.testing.assert_almost_equal(po_vars[0]["v1"], 0.0)
        assert po_vars[0]["v2"] == 1.0

    @pytest.mark.parametrize(
        "inner_solver",
        [
            GridSearchSolver(GridSearchSolver.Params(n_grids_per_var=[2, 7])),
            RandomSamplerSolver(RandomSamplerSolver.Params(n_samples_per_param=1000)),
        ],
    )
    def test_solver_with_two_obj_problem_with_cache(
        self, inner_solver: SOSolver, two_obj_problem: MOProblem
    ) -> None:
        ws_pairs = np.array(
            [
                [0.3, 0.7],
                [0.6, 0.4],
                [0.1, 0.9],
                [0.2, 0.8],
                [0.4, 0.6],
                [0.5, 0.5],
            ]
        )

        ws_algo = WeightedSum(
            WeightedSum.Params(
                so_solver=inner_solver,
                ws_pairs=ws_pairs,
                allow_cache=True,
            )
        )
        po_objs, po_vars = ws_algo.solve(problem=two_obj_problem, seed=0)

        np.testing.assert_almost_equal(po_objs, np.array([[0, 0]]), decimal=3)
        np.testing.assert_almost_equal(po_vars[0]["v1"], 0.0, decimal=2)
        assert po_vars[0]["v2"] == 1.0

    @pytest.mark.parametrize(
        "inner_solver",
        [
            GridSearchSolver(GridSearchSolver.Params(n_grids_per_var=[2, 7])),
            RandomSamplerSolver(RandomSamplerSolver.Params(n_samples_per_param=10)),
        ],
    )
    def test_ws_raises_no_solution(
        self, inner_solver: SOSolver, simple_problem: MOProblem
    ) -> None:
        ws_pairs = np.array([[0.3, 0.7], [0.6, 0.4]])

        def f3(
            input_variables: Dict[str, th.Tensor],
            input_parameters: Optional[Dict[str, th.Tensor]] = None,
        ) -> th.Tensor:
            return input_variables["v1"] + input_variables["v2"] - 10

        simple_problem.constraints = [Constraint(function=f3, lower=0)]
        simple_problem.input_parameters = None
        ws_algo = WeightedSum(
            WeightedSum.Params(
                so_solver=inner_solver,
                ws_pairs=ws_pairs,
            )
        )
        with pytest.raises(NoSolutionError):
            ws_algo.solve(problem=simple_problem, seed=0)

    @pytest.mark.parametrize(
        "inner_solver",
        [
            GridSearchSolver(GridSearchSolver.Params(n_grids_per_var=[2, 7])),
            RandomSamplerSolver(RandomSamplerSolver.Params(n_samples_per_param=30)),
        ],
    )
    def test_works_with_three_objectives(
        self, inner_solver: SOSolver, simple_problem: MOProblem
    ) -> None:
        ws_pairs = np.array([[0.3, 0.5, 0.2], [0.6, 0.3, 0.1]])

        def f2(
            input_variables: Dict[str, th.Tensor],
            input_parameters: Optional[Dict[str, th.Tensor]] = None,
        ) -> th.Tensor:
            return input_variables["v2"]

        objectives = list(simple_problem.objectives)
        objectives.insert(1, Objective("obj3", function=f2, minimize=True))
        simple_problem.objectives = objectives
        simple_problem.input_parameters = None

        def constraint_f(
            input_variables: Dict[str, th.Tensor],
            input_parameters: Optional[Dict[str, th.Tensor]] = None,
        ) -> th.Tensor:
            return input_variables["v1"] + input_variables["v2"] - 3

        simple_problem.constraints = [Constraint(function=constraint_f, lower=0)]
        ws_algo = WeightedSum(
            WeightedSum.Params(
                so_solver=inner_solver,
                ws_pairs=ws_pairs,
            )
        )
        po_objs, po_vars = ws_algo.solve(problem=simple_problem, seed=0)

        np.testing.assert_almost_equal(po_objs, np.array([[0, 3, 0.3]]))
        np.testing.assert_equal(po_vars, np.array([{"v1": 0, "v2": 3}]))



from typing import Tuple

import dgl
import pytest
import torch as th

from ....embedders.layers.multi_head_attention import (
    MultiHeadAttentionLayer,
    QFMultiHeadAttentionLayer,
    RAALMultiHeadAttentionLayer,
)
from ....utils.utils import set_deterministic_torch

FixtureType = Tuple[MultiHeadAttentionLayer, dgl.DGLGraph, th.Tensor]


@pytest.fixture
def mha_fixture() -> FixtureType:
    g = dgl.graph(([0, 1, 2], [1, 2, 3]))
    h = th.rand((4, 5))
    mha = MultiHeadAttentionLayer(in_dim=5, out_dim=3, n_heads=2, use_bias=True)

    return mha, g, h


class TestMultiHeadAttentionLayer:
    def test_init(self) -> None:
        mha = MultiHeadAttentionLayer(in_dim=5, out_dim=3, n_heads=2, use_bias=True)
        assert mha.Q.in_features == 5
        assert mha.Q.out_features == 6
        assert mha.K.in_features == 5
        assert mha.K.out_features == 6
        assert mha.V.in_features == 5
        assert mha.V.out_features == 6

    def test_forward(self, mha_fixture: FixtureType) -> None:
        set_deterministic_torch(0)
        mha, g, h = mha_fixture
        out = mha(g, h)
        expected = th.tensor(
            [
                [
                    [0.000000e00, 0.000000e00, 0.000000e00],
                    [0.000000e00, 0.000000e00, 0.000000e00],
                ],
                [
                    [-5.366904e-01, 6.547331e-02, -7.822125e-02],
                    [1.462809e-01, -3.937286e-01, -7.183467e-01],
                ],
                [
                    [-5.425442e-01, -1.044514e-02, -1.592951e-01],
                    [-1.745791e-01, -5.798197e-01, -6.325653e-01],
                ],
                [
                    [-7.297680e-01, 9.695685e-02, -3.332692e-01],
                    [-6.688768e-03, -5.367336e-01, -7.892159e-01],
                ],
            ]
        )
        assert out.size() == (4, 2, 3)
        assert th.allclose(out[1:], expected[1:], rtol=1e-5)

    def test_compute_query_key_value(self, mha_fixture: FixtureType) -> None:
        mha, g, h = mha_fixture
        g = mha.compute_query_key_value(g, h)
        for key in ["Q_h", "K_h", "V_h"]:
            assert key in g.ndata
            assert g.ndata[key].size() == (4, 2, 3)  # type: ignore
        pass

    def test_compute_attention(self, mha_fixture: FixtureType) -> None:
        mha, g, h = mha_fixture
        graph = mha.compute_query_key_value(g, h)
        g = mha.compute_attention(graph)
        assert "score" in g.edata

    def test_attention_has_effect(self, mha_fixture: FixtureType) -> None:
        mha, g, h = mha_fixture
        out = mha(g, h)
        original_h = h.clone()
        assert not th.equal(original_h, out[:, 0, :])


class TestQFMultiHeadAttentionLayer:
    def test_compute_attention(self) -> None:
        set_deterministic_torch(0)
        g = dgl.graph(([0, 1, 2], [1, 2, 3]))
        h = th.rand((4, 5))
        g.edata["dist"] = th.tensor([1, 2, 3])
        attention_bias = th.tensor([0.1, 0.2, 0.3])
        mha = QFMultiHeadAttentionLayer(
            in_dim=5,
            out_dim=3,
            n_heads=2,
            use_bias=True,
            attention_bias=attention_bias,
        )
        g = mha.compute_query_key_value(g, h)
        g = mha.compute_attention(g)
        expected_score = th.tensor(
            [
                [[9.535783e-01], [1.067863e00]],
                [[1.099911e00], [1.133566e00]],
                [[1.203815e00], [1.290228e00]],
            ]
        )

        assert "score" in g.edata
        assert th.allclose(g.edata["score"], expected_score, rtol=1e-5)  # type: ignore


class TestRAALMultiHeadAttentionLayer:
    def test_compute_attention(self) -> None:
        set_deterministic_torch(0)
        g = dgl.graph(([0, 1, 2], [1, 2, 3]))
        g.ndata["sid"] = th.tensor([0, 0, 0, 0])
        non_siblings_map = {0: {0: [2, 3], 1: [0, 3], 2: [0, 1]}}
        h = th.rand((4, 5))
        mha = RAALMultiHeadAttentionLayer(
            in_dim=5,
            out_dim=3,
            n_heads=2,
            use_bias=True,
            non_siblings_map=non_siblings_map,
        )
        g = mha.compute_query_key_value(g, h)
        g = mha.compute_attention(g)
        expected_score = th.tensor(
            [
                [[3.676294e-01], [8.327479e-01]],
                [[2.522929e-01], [3.981344e00]],
                [[6.267625e-01], [3.711438e-01]],
            ]
        )
        assert "score" in g.edata
        assert th.allclose(g.edata["score"], expected_score, rtol=1e-5)  # type: ignore



from .embedders import BaseEmbedder, BaseGraphEmbedder, GraphAverager, GraphTransformer
from .model import FixedEmbeddingUdaoModel, UdaoModel
from .module import UdaoModule
from .regressors import MLP, BaseRegressor
from .utils import losses, metrics, schedulers, utils

__all__ = [
    "BaseEmbedder",
    "BaseGraphEmbedder",
    "BaseRegressor",
    "GraphAverager",
    "GraphTransformer",
    "MLP",
    "UdaoModel",
    "FixedEmbeddingUdaoModel",
    "UdaoModule",
    "losses",
    "metrics",
    "schedulers",
    "utils",
]



from typing import Dict, Iterable, Optional

import numpy as np
import pytest
import torch as th

from ... import concepts as co
from ...soo.grid_search_solver import GridSearchSolver


class TestGridSearch:
    @pytest.mark.parametrize(
        "test_data, expected",
        [
            ({"variable": co.BoolVariable(), "n_grids": 1}, [0]),
            ({"variable": co.BoolVariable(), "n_grids": 2}, [0, 1]),
            ({"variable": co.BoolVariable(), "n_grids": 3}, [0, 1]),
            (
                {"variable": co.IntegerVariable(1, 7), "n_grids": 5},
                [1, 2, 4, 6, 7],
            ),
            (
                {"variable": co.IntegerVariable(1, 7), "n_grids": 8},
                [1, 2, 3, 4, 5, 6, 7],
            ),
            (
                {"variable": co.FloatVariable(2, 4), "n_grids": 5},
                [2, 2.5, 3, 3.5, 4],
            ),
            (
                {"variable": co.EnumVariable([0, 4, 7, 10]), "n_grids": 2},
                [0, 4, 7, 10],
            ),
        ],
    )
    def test_grid_search_get_single_variable(
        self, test_data: Dict, expected: Iterable
    ) -> None:
        solver = GridSearchSolver(
            GridSearchSolver.Params(n_grids_per_var=[test_data["n_grids"]])
        )
        output = solver._get_input(
            variables={"variable": test_data["variable"]},
        )
        np.testing.assert_equal(output, {"variable": np.array([e for e in expected])})

    def test_grid_search_get_multiple_variables(self) -> None:
        solver = GridSearchSolver(GridSearchSolver.Params(n_grids_per_var=[2, 7]))
        output = solver._get_input(
            variables={"v1": co.BoolVariable(), "v2": co.IntegerVariable(1, 7)},
        )
        expected_array = np.array(
            [
                [0, 1],
                [0, 2],
                [0, 3],
                [0, 4],
                [0, 5],
                [0, 6],
                [0, 7],
                [1, 1],
                [1, 2],
                [1, 3],
                [1, 4],
                [1, 5],
                [1, 6],
                [1, 7],
            ]
        ).T
        np.testing.assert_equal(
            output, {"v1": expected_array[0], "v2": expected_array[1]}
        )

    def test_solve(self) -> None:
        solver = GridSearchSolver(GridSearchSolver.Params(n_grids_per_var=[2, 7]))

        def obj1_func(
            input_variables: Dict[str, th.Tensor],
            input_parameters: Optional[Dict[str, th.Tensor]] = None,
        ) -> th.Tensor:
            return input_variables["v1"] + input_variables["v2"]

        objective = co.Objective("obj1", minimize=False, function=obj1_func)
        variables: Dict[str, co.Variable] = {
            "v1": co.BoolVariable(),
            "v2": co.IntegerVariable(1, 7),
        }
        problem = co.SOProblem(objective=objective, variables=variables, constraints=[])
        soo_obj, soo_vars = solver.solve(problem)
        assert soo_obj == 8
        assert soo_vars == {"v1": 1, "v2": 7}



from dataclasses import dataclass

import torch as th

from ...regressors.base_regressor import BaseRegressor


class DummyRegressor(BaseRegressor):
    @dataclass
    class Params(BaseRegressor.Params):
        pass

    def __init__(self, net_params: Params) -> None:
        super().__init__(net_params)

    def forward(self, embedding: th.Tensor, inst_feat: th.Tensor) -> th.Tensor:
        input_vector = th.cat([embedding, inst_feat], dim=1)
        return th.sum(input_vector, dim=1)



from abc import ABC, abstractmethod
from typing import Any, Optional

from ..concepts.problem import MOProblem


class MOSolver(ABC):
    def __init__(
        self,
    ) -> None:
        pass

    @abstractmethod
    def solve(
        self,
        problem: MOProblem,
        seed: Optional[int] = None,
    ) -> Any:
        """_summary_

        Parameters
        ----------
        problem : MOProblem
            Multi-objective optimization problem to solve
        seed : Optional[int], optional
            Random seed, by default None

        Returns
        -------
        Any
            A tuple of the objectives values and the variables
            that optimize the objective
        """
        ...



from .base_preprocessor import (
    FeaturePreprocessor,
    StaticPreprocessor,
    TrainedPreprocessor,
)
from .normalize_preprocessor import NormalizePreprocessor
from .one_hot_preprocessor import OneHotPreprocessor

__all__ = [
    "FeaturePreprocessor",
    "NormalizePreprocessor",
    "OneHotPreprocessor",
    "StaticPreprocessor",
    "TrainedPreprocessor",
]



from typing import Callable, Dict, List, Tuple

import pandas as pd

from ..containers import TabularContainer
from ..predicate_embedders import BasePredicateEmbedder
from ..predicate_embedders.utils import extract_operations, prepare_operation
from .base_extractors import TrainedExtractor


class PredicateEmbeddingExtractor(TrainedExtractor[TabularContainer]):
    """Class to extract embeddings from a DataFrame of query plans.

    Parameters
    ----------
    embedder : BaseEmbedder
        Embedder to use to extract the embeddings,
        e.g. an instance of Word2Vecembedder.
    """

    def __init__(
        self,
        embedder: BasePredicateEmbedder,
        op_preprocessing: Callable[[str], str] = prepare_operation,
        extract_operations: Callable[
            [pd.DataFrame, Callable], Tuple[Dict[int, List[int]], List[str]]
        ] = extract_operations,
    ) -> None:
        self.embedder = embedder
        self.op_preprocessing = op_preprocessing
        self.extract_operations = extract_operations

    def extract_features(self, df: pd.DataFrame, split: str) -> TabularContainer:
        """Extract embeddings from a DataFrame of query plans.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame containing the query plans and their ids.
        split : str
            Split of the dataset, either "train", "test" or "validation".
            Will fit the embedder if "train" and transform otherwise.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the embeddings of each operation of the query plans.
        """

        plan_to_operations, operations_list = self.extract_operations(
            df, self.op_preprocessing
        )
        if split == "train":
            embeddings_list = self.embedder.fit_transform(operations_list)
        else:
            embeddings_list = self.embedder.transform(operations_list)
        emb_series = df["id"].apply(
            lambda idx: [embeddings_list[op_id] for op_id in plan_to_operations[idx]]  # type: ignore
        )
        emb_df = emb_series.to_frame("embeddings")
        emb_df["plan_id"] = df["id"]
        emb_df = emb_df.explode("embeddings", ignore_index=True)
        embedding_length = len(emb_df["embeddings"].iloc[0])
        emb_df[[f"emb_{i}" for i in range(embedding_length)]] = pd.DataFrame(
            emb_df.embeddings.tolist(),
            index=emb_df.index,
        )
        emb_df = emb_df.drop(columns=["embeddings"])
        emb_df["operation_id"] = emb_df.groupby("plan_id").cumcount()
        emb_df = emb_df.set_index(["plan_id", "operation_id"])
        return TabularContainer(emb_df)



from typing import Dict, List

import torch as th
from torchmetrics import Metric

from .utils import EPS


class ErrorPercentiles(Metric):
    def __init__(
        self,
        percentiles: List[float] = [50, 90, 95, 99],
        dist_sync_on_step: bool = False,
    ) -> None:
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.quantiles = th.tensor(percentiles) / 100
        self.add_state("error_rate", default=th.empty((0)), dist_reduce_fx=None)

    def update(self, input: th.Tensor, target: th.Tensor) -> None:
        new_error_rate = th.abs(input - target) / (th.clip(target, min=EPS))
        self.error_rate = th.cat([self.error_rate, new_error_rate])  # type: ignore

    def compute(self) -> Dict[str, th.Tensor]:
        error_quantiles = th.quantile(self.error_rate, self.quantiles, dim=0)  # type: ignore
        return {
            "mean": self.error_rate.mean(),  # type: ignore
            **{
                f"q_{int(q * 100)}": error_quantiles[i]
                for i, q in enumerate(self.quantiles)
            },
        }


class QErrorPercentiles(Metric):
    def __init__(
        self,
        percentiles: List[int] = [50, 90, 95, 99],
        dist_sync_on_step: bool = False,
    ) -> None:
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.quantiles = th.tensor(percentiles) / 100
        self.add_state("q_error", default=th.empty((0)), dist_reduce_fx=None)

    def update(self, input: th.Tensor, target: th.Tensor) -> None:
        new_q_error = th.maximum(
            th.clip(target, EPS) / th.clip(input, EPS),
            th.clip(input, EPS) / th.clip(target, EPS),
        )
        self.q_error = th.cat([self.q_error, new_q_error])  # type: ignore

    def compute(self) -> Dict[str, th.Tensor]:
        q_error_quantiles = th.quantile(self.q_error, self.quantiles, dim=0)  # type: ignore
        return {
            "mean": self.q_error.mean(),  # type: ignore
            **{
                f"q_{int(q * 100)}": q_error_quantiles[i]
                for i, q in enumerate(self.quantiles)
            },
        }



from typing import Any

import pytest
import torch as th
from torch import nn
from torchmetrics import WeightedMeanAbsolutePercentageError

from ..module import UdaoModule
from ..utils.losses import WMAPELoss
from ..utils.utils import set_deterministic_torch


@pytest.fixture
def sample_module() -> UdaoModule:
    set_deterministic_torch(0)

    model = nn.Linear(2, 2)
    objectives = ["obj1", "obj2"]
    loss = WMAPELoss()
    module = UdaoModule(
        model=model,
        objectives=objectives,
        loss=loss,
        metrics=[WeightedMeanAbsolutePercentageError],
    )
    return module


class TestUdaoModule:
    def test_initialize(self, sample_module: UdaoModule) -> None:
        assert sample_module.loss_weights == {"obj1": 1.0, "obj2": 1.0}
        assert isinstance(sample_module.optimizer, th.optim.AdamW)
        # Model, loss and metrics should be children of the module
        assert len([child for child in sample_module.children()]) == 3

    def test_compute_loss(self, sample_module: UdaoModule) -> None:
        y = th.tensor([[1, 1], [2, 2]], dtype=th.float32)
        y_hat = th.tensor([[2, 2], [1, 1]], dtype=th.float32)
        loss, loss_dict = sample_module.compute_loss(y, y_hat)
        assert th.allclose(loss, th.tensor(4 / 3))
        assert loss_dict == {"obj1": 2 / 3, "obj2": 2 / 3}

    def test_training_step(self, sample_module: UdaoModule, mocker: Any) -> None:
        batch = (
            th.tensor([[1, 1], [2, 2]], dtype=th.float32),
            th.tensor([[2, 2], [1, 1]], dtype=th.float32),
        )
        mocked_log = mocker.patch.object(sample_module, "log")
        returned = sample_module.training_step(batch, 0)
        assert mocked_log.called
        assert th.allclose(returned, th.tensor(2.7835))

    def test__shared_epoch_end(self, sample_module: UdaoModule, mocker: Any) -> None:
        batch = (
            th.tensor([[1, 1], [2, 2]], dtype=th.float32),
            th.tensor([[2, 2], [1, 1]], dtype=th.float32),
        )

        mocked_log = mocker.spy(sample_module, "log")
        loss = sample_module.training_step(batch, 0)

        sample_module._shared_epoch_end("train")
        # assert the metrics are logged once per objective
        assert mocked_log.call_count == 3
        logged_loss = mocked_log.call_args_list[0][0]
        assert logged_loss[0] == "train_loss"
        obj1_metric = mocked_log.call_args_list[1][0]
        assert obj1_metric[0] == "train_obj1_WeightedMeanAbsolutePercentageError"
        obj2_metric = mocked_log.call_args_list[2][0]
        assert obj2_metric[0] == "train_obj2_WeightedMeanAbsolutePercentageError"

        # assert the loss is the same as the metrics (WMAPE in both cases)
        assert th.equal(obj1_metric[1] + obj2_metric[1], loss)
        assert th.allclose(
            obj1_metric[1],
            th.tensor(0.8075),
            rtol=1e-4,
        )
        assert th.allclose(
            obj2_metric[1],
            th.tensor(1.9760),
            rtol=1e-4,
        )
        assert th.equal(logged_loss[1], loss)

    def test_shared_step(self, sample_module: UdaoModule) -> None:
        batch = (
            th.tensor([[0.5, 2], [0.5, 2]], dtype=th.float32),
            th.tensor([[1, 1], [1, 1]], dtype=th.float32),
        )
        y_hat, y = sample_module._shared_step(batch, "train")
        assert y_hat.shape == (2, 2)
        assert y.shape == (2, 2)



from .concepts import (
    BoolVariable,
    Constraint,
    EnumVariable,
    FloatVariable,
    IntegerVariable,
    MOProblem,
    NumericVariable,
    Objective,
    SOProblem,
    Variable,
)
from .moo.progressive_frontier import (
    ParallelProgressiveFrontier,
    SequentialProgressiveFrontier,
)
from .moo.weighted_sum import WeightedSum
from .soo.grid_search_solver import GridSearchSolver
from .soo.mogd import MOGD
from .soo.random_sampler_solver import RandomSamplerSolver

__all__ = [
    "Constraint",
    "Objective",
    "Variable",
    "NumericVariable",
    "IntegerVariable",
    "FloatVariable",
    "EnumVariable",
    "BoolVariable",
    "MOProblem",
    "SOProblem",
    "SequentialProgressiveFrontier",
    "ParallelProgressiveFrontier",
    "WeightedSum",
    "GridSearchSolver",
    "RandomSamplerSolver",
    "MOGD",
]



from collections import defaultdict
from typing import Any, Callable, Dict, List, Literal, Tuple, Type

import dgl
import dgl.function as fn
import numpy as np
import torch
import torch.nn as nn


def e_src_dot_dst(
    src_field: str, dst_field: str, out_field: str
) -> Callable[[Any], Dict[str, torch.Tensor]]:
    """Multiply source and destination node features and sum them up"""

    def func(edges: Any) -> Dict[str, torch.Tensor]:
        return {
            out_field: (edges.src[src_field] * edges.dst[dst_field]).sum(
                -1, keepdim=True
            )
        }

    return func


def e_scaled_exp(
    field: str, scale_constant: float
) -> Callable[[Any], Dict[str, torch.Tensor]]:
    """Compute scaled exponential of a graph's edge field"""

    def func(edges: Any) -> Dict[str, torch.Tensor]:
        # clamp for softmax numerical stability
        return {field: torch.exp((edges.data[field] / scale_constant).clamp(-5, 5))}

    return func


class MultiHeadAttentionLayer(nn.Module):
    """Multi-Head Attention Layer for Graph
    proposed by "A Generalization of Transformer Networks to Graphs", DLG-AAAI'21.
    https://arxiv.org/pdf/2012.09699.pdf

    Parameters
    ----------
        in_dim : int
            Input dimension
        out_dim : int
            Output dimension
        n_heads : int
            Number of attention heads
        use_bias : bool
            Whether to use bias
    """

    def __init__(self, in_dim: int, out_dim: int, n_heads: int, use_bias: bool) -> None:
        super().__init__()
        self.out_dim = out_dim
        self.n_heads = n_heads

        self.Q = nn.Linear(in_dim, out_dim * n_heads, bias=use_bias)
        self.K = nn.Linear(in_dim, out_dim * n_heads, bias=use_bias)
        self.V = nn.Linear(in_dim, out_dim * n_heads, bias=use_bias)

    def _apply_attention(self, g: dgl.DGLGraph) -> torch.Tensor:
        edge_ids: Tuple[torch.Tensor, torch.Tensor] = g.edges()
        g.send_and_recv(
            edge_ids,
            fn.u_mul_e("V_h", "score", "V_h"),  # type: ignore
            fn.sum("V_h", "wV"),  # type: ignore
        )
        g.send_and_recv(
            edge_ids, fn.copy_e("score", "score"), fn.sum("score", "z")  # type: ignore
        )
        head_out = g.ndata["wV"] / (
            g.ndata["z"] + torch.full_like(g.ndata["z"], 1e-6)  # type: ignore
        )
        # uncommenting below would delete intermediate values from graph.
        # Necessary if hidden_dim is different from out_dim
        # g.ndata.pop("wV")
        # g.ndata.pop("z")
        # g.edata.pop("score")
        return head_out

    def compute_attention(self, g: dgl.DGLGraph) -> dgl.DGLGraph:
        """Simple attention mechanism"""
        g.apply_edges(e_src_dot_dst("K_h", "Q_h", "score"))
        g.apply_edges(e_scaled_exp("score", np.sqrt(self.out_dim)))
        return g

    def compute_query_key_value(self, g: dgl.DGLGraph, h: torch.Tensor) -> dgl.DGLGraph:
        """Compute query, key, and value for each node"""
        g.ndata["Q_h"] = self.Q(h).view(-1, self.n_heads, self.out_dim)
        g.ndata["K_h"] = self.K(h).view(-1, self.n_heads, self.out_dim)
        g.ndata["V_h"] = self.V(h).view(-1, self.n_heads, self.out_dim)
        return g

    def forward(self, g: dgl.DGLGraph, h: torch.Tensor) -> torch.Tensor:
        g = self.compute_query_key_value(g, h)
        g = self.compute_attention(g)
        return self._apply_attention(g)


class RAALMultiHeadAttentionLayer(MultiHeadAttentionLayer):
    """MultiHead Attention using Resource-Aware Attentional LSTM
    proposed by "A Resource-Aware Deep Cost Model for Big Data Query Processing”
    https://ieeexplore.ieee.org/document/9835426

    The RAAL MultiHead Attention Layer requires the graphs to have an "sid" node
    feature, corresponding to the id of the template graph.
    This makes the link with the non_siblings_map.

    Parameters
    ----------
    non_siblings_map : Dict[int, Dict[int, List[int]]]
    For each type of graph, maps the edge id to
    all nodes that are not siblings of its source node
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        n_heads: int,
        use_bias: bool,
        non_siblings_map: Dict[int, Dict[int, List[int]]],
    ) -> None:
        super().__init__(in_dim, out_dim, n_heads, use_bias)
        self.non_siblings_map = non_siblings_map

    def compute_attention(self, g: dgl.DGLGraph) -> dgl.DGLGraph:
        """Attention mechanism with non-siblings attention"""
        g_list = dgl.unbatch(g)
        sid_g_map: Dict[int, List[dgl.DGLGraph]] = defaultdict(list)
        for gg in g_list:
            sid = gg.ndata["sid"][0].cpu().item()
            sid_g_map[sid].append(gg)

        gb_list = []
        for sid, graphs in sid_g_map.items():
            n_graphs = len(graphs)
            gb = dgl.batch(graphs)
            Q = gb.ndata["Q_h"].reshape(n_graphs, -1, self.n_heads, self.out_dim)  # type: ignore
            K = gb.ndata["K_h"].reshape(n_graphs, -1, self.n_heads, self.out_dim)  # type: ignore
            QK = (
                torch.matmul(Q.transpose(1, 2), K.transpose(1, 2).transpose(2, 3))
                .transpose(1, 2)
                .clamp(-5, 5)
            )
            srcs, dsts, eids = graphs[0].edges(form="all", order="srcdst")
            score_list = [
                QK[:, src, :, dst]
                / (
                    QK[:, src, :, self.non_siblings_map[sid][eid]].sum(-1)
                    + torch.full_like(QK[:, src, :, dst], 1e-6)
                )
                for src, dst, eid in zip(
                    srcs.cpu().numpy(), dsts.cpu().numpy(), eids.cpu().numpy()
                )
            ]
            gb.edata["score"] = torch.cat(score_list, dim=1).view(-1, self.n_heads, 1)
            gb_list.append(gb)
        return dgl.batch(gb_list)


class QFMultiHeadAttentionLayer(MultiHeadAttentionLayer):
    """MultiHead Attention using QueryFormer
    proposed by "QueryFormer: A Tree Transformer Model for Query Plan
    Representation"
    https://www.vldb.org/pvldb/vol15/p1658-zhao.pdf

    The QF MultiHead Attention Layer requires the graphs to have a "dist"
    edge feature. see examples/data/spark/4.qf_addition

    Parameters
    ----------
    attention_bias : torch.Tensor
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        n_heads: int,
        use_bias: bool,
        attention_bias: torch.Tensor,
    ) -> None:
        super().__init__(in_dim, out_dim, n_heads, use_bias)
        self.attention_bias = attention_bias

    def compute_attention(self, g: dgl.DGLGraph) -> dgl.DGLGraph:
        """Attention mechanism with attention bias"""
        g = super().compute_attention(g)
        g.edata["att_bias"] = torch.index_select(
            self.attention_bias, 0, g.edata["dist"] - 1  # type: ignore
        ).reshape(-1, 1, 1)
        g.edata["score"] = g.edata["score"] + g.edata["att_bias"]  # type: ignore
        return g


AttentionLayerName = Literal["QF", "GTN", "RAAL"]

ATTENTION_TYPES: Dict[AttentionLayerName, Type[MultiHeadAttentionLayer]] = {
    "QF": QFMultiHeadAttentionLayer,
    "GTN": MultiHeadAttentionLayer,
    "RAAL": RAALMultiHeadAttentionLayer,
}



import random

import numpy as np
import torch

EPS = 1e-7


def set_deterministic_torch(seed: int = 0) -> None:
    """
    Set seeds and configurations to enable deterministic behavior in PyTorch.

    Parameters
    ----------
    seed : int
        Random seed to use.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.use_deterministic_algorithms(True)  # type: ignore



import torch as th

from ...utils.losses import WMAPELoss


def test_wmape_loss() -> None:
    y = th.tensor([[1, 1, 1, 1, 1], [2, 2, 2, 2, 2]])
    y_hat = th.tensor([[2, 2, 2, 2, 2], [1, 1, 1, 1, 1]])
    loss = WMAPELoss()

    assert loss(y_hat, y) == 2 / 3



import numpy as np
import pytest

from ...predicate_embedders import Word2VecEmbedder, Word2VecParams


@pytest.fixture
def word2vec_embedder() -> Word2VecEmbedder:
    return Word2VecEmbedder(Word2VecParams())


class TestWord2VecEmbedder:
    def test_init(self, word2vec_embedder: Word2VecEmbedder) -> None:
        assert word2vec_embedder.w2v_model is not None
        assert word2vec_embedder._bigram_model is None

    def test_fit_transform(self, word2vec_embedder: Word2VecEmbedder) -> None:
        training_plans = ["a b c", "a b d"]
        training_encodings = word2vec_embedder.fit_transform(training_plans)
        assert word2vec_embedder._bigram_model is not None
        # 4 words
        assert word2vec_embedder.w2v_model.wv.vectors.shape[0] == 4
        # 32 dimensions - corresponds to param vec_size
        assert word2vec_embedder.w2v_model.wv.vectors.shape[1] == 32
        # 2 training plans with dimension 32
        assert training_encodings.shape == (2, 32)

    def test_transform_not_trained(self, word2vec_embedder: Word2VecEmbedder) -> None:
        with pytest.raises(ValueError):
            word2vec_embedder.transform(["a b c"])

    def test_transform_trained(self, word2vec_embedder: Word2VecEmbedder) -> None:
        training_plans = ["a b c", "a b d"]
        training_encodings = word2vec_embedder.fit_transform(training_plans)
        encoding = word2vec_embedder.transform(["a b c", "a b x"])
        assert np.array_equal(training_encodings[0], encoding[0])
        assert encoding.shape == (2, 32)



from dataclasses import dataclass
from enum import Enum
from typing import Generic, TypeVar

import torch as th

T = TypeVar("T")

ST = TypeVar("ST")


@dataclass
class UdaoInput:
    features: th.Tensor

    def to(self, device: th.device) -> "UdaoInput":
        return UdaoInput(self.features.to(device))


@dataclass
class UdaoItemShape:
    feature_names: list[str]
    output_names: list[str]


@dataclass
class UdaoEmbedInput(Generic[T], UdaoInput):
    embedding_input: T

    def to(self, device: th.device) -> "UdaoEmbedInput":
        if hasattr(self.embedding_input, "to"):
            return UdaoEmbedInput(
                self.embedding_input.to(device), self.features.to(device)  # type: ignore
            )
        else:
            return UdaoEmbedInput(
                self.embedding_input, self.features.to(device)  # type: ignore
            )


@dataclass
class UdaoEmbedItemShape(Generic[ST], UdaoItemShape):
    embedding_input_shape: ST


class VarTypes(Enum):
    INT = "int"
    BOOL = "bool"
    CATEGORY = "category"
    FLOAT = "float"



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Union

from ..containers.base_container import BaseContainer
from ..utils.utils import DatasetType

T = TypeVar("T", bound=BaseContainer)


class TrainedPreprocessor(ABC, Generic[T]):
    """Base class for feature processors that require training."""

    trained: bool = True

    def __init__(self) -> None:
        pass

    @abstractmethod
    def preprocess(self, container: T, split: DatasetType) -> T:
        pass


class StaticPreprocessor(ABC, Generic[T]):
    """Base class for feature processors that do not require training."""

    trained: bool = False

    def __init__(self) -> None:
        pass

    @abstractmethod
    def preprocess(self, container: T) -> T:
        pass


FeaturePreprocessor = Union[TrainedPreprocessor, StaticPreprocessor]



from typing import Dict, Optional, Sequence, Tuple

import numpy as np
import torch as th
from matplotlib import pyplot as plt

from .exceptions import NoSolutionError


class Point:
    def __init__(self, objs: np.ndarray, vars: Optional[Dict] = None) -> None:
        """
        A point in the objective space.
        Variables are optional, and are not specified for imaginary points
        (e.g., utopia and nadir)

        Parameters
        ----------
        objs : np.ndarray
            Array of objective values of shape (n_objs,)
        vars :np.ndarray, optional
            Array of variable values of shape (n_vars,), by default None
        """
        self.objs = objs
        self.vars = vars
        self.n_objs = objs.shape[0]

    def __repr__(self) -> str:
        return f"Point(objs={self.objs}, vars={self.vars})"

    def __eq__(self, other: "Point") -> bool:  # type: ignore
        return bool(np.all(self.objs == other.objs) and np.all(self.vars == other.vars))


class Rectangle:
    def __init__(self, utopia: Point, nadir: Point) -> None:
        """

        Parameters
        ----------
        utopia : Points
            utopia point
        nadir : Points
            nadir point
        """

        self.upper_bounds = nadir.objs
        self.lower_bounds = utopia.objs
        self.n_objs = nadir.objs.shape[0]
        self.volume = self.cal_volume(nadir.objs, utopia.objs)
        self.neg_vol = -self.volume
        self.utopia = utopia
        self.nadir = nadir

    def __repr__(self) -> str:
        return f"Rectangle(utopia={self.utopia}, nadir={self.nadir})"

    def cal_volume(self, upper_bounds: np.ndarray, lower_bounds: np.ndarray) -> float:
        """
        Calculate the volume of the hyper_rectangle

        Parameters
        ----------
        upper_bounds : np.ndarray(
            Array of upper bounds of the hyper_rectangle, of shape (n_objs,)
        lower_bounds : np.ndarrays
            Array of lower bounds of the hyper_rectangle of shape (n_objs,)

        Returns
        -------
        float
            volume of the hyper_rectangle
        """
        volume = np.abs(np.prod(upper_bounds - lower_bounds))
        return volume

    # Override the `__lt__()` function to make `Rectangles`
    # class work with min-heap (referred from VLDB2022)
    def __lt__(self, other: "Rectangle") -> bool:
        return self.neg_vol < other.neg_vol

    def __eq__(self, other: "Rectangle") -> bool:  # type: ignore
        return bool(
            np.all(self.upper_bounds == other.upper_bounds)
            and np.all(self.lower_bounds == other.lower_bounds)
        )


# a quite efficient way to get the indexes of pareto points
# https://stackoverflow.com/a/40239615
def is_pareto_efficient(costs: np.ndarray, return_mask: bool = True) -> np.ndarray:
    ## reuse code in VLDB2022
    """
    Find the pareto-efficient points

    Parameters
    ----------
    costs : np.ndarray
        An (n_points, n_costs) array
    return_mask : bool, default=True
        True to return a mask

    Returns
    -------
    np.ndarray
        If return_mask is True, this will be an (n_points, ) boolean array
        Otherwise it will be a (n_efficient_points, ) integer array of indices.
    """
    is_efficient = np.arange(costs.shape[0])
    n_points = costs.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs):
        nondominated_point_mask = np.any(costs < costs[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        costs = costs[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1
    if return_mask:
        is_efficient_mask = np.zeros(n_points, dtype=bool)
        is_efficient_mask[is_efficient] = True
        return is_efficient_mask
    else:
        return is_efficient


def summarize_ret(
    po_obj_list: Sequence, po_var_list: Sequence
) -> Tuple[np.ndarray, np.ndarray]:
    """Return the pareto-optimal objectives and variables

    Parameters
    ----------
    po_obj_list: Sequence
        List of objective values
    po_var_list : _type_
        List of variable values

    Returns
    -------
    Tuple[Optional[np.ndarray], Optional[np.ndarray]]
        Pareto-optimal objectives and variables
    """
    ## reuse code in VLDB2022
    if len(po_obj_list) == 0:
        raise NoSolutionError("No feasible solutions found: empty po_obj_list")
    elif len(po_obj_list) == 1:
        return np.array(po_obj_list), np.array(po_var_list)
    else:
        po_objs_cand = np.array(po_obj_list)
        po_vars_cand = np.array(po_var_list)
        po_inds = is_pareto_efficient(po_objs_cand)
        po_objs = po_objs_cand[po_inds]
        po_vars = po_vars_cand[po_inds]

        return po_objs, po_vars


# generate even weights for 2d and 3D
def even_weights(stepsize: float, n_objectives: int) -> np.ndarray:
    """Generate even weights for 2d and 3D

    Parameters
    ----------
    stepsize : float
        Step size for the weights
    n_objectives : int
        Number of objectives for which to generate weights
        Only 2 and 3 are supported

    Returns
    -------
    np.ndarray
        Array of weights of shape (n_weights, n_objectives)

    Raises
    ------
    Exception
        If `n_objectives` is not 2 or 3
    """
    ws_pairs = np.array([])
    if n_objectives == 2:
        w1 = np.hstack([np.arange(0, 1, stepsize), 1])
        w2 = 1 - w1
        ws_pairs = np.array([[w1, w2] for w1, w2 in zip(w1, w2)])

    elif n_objectives == 3:
        w_steps = np.linspace(0, 1, num=int(1 / stepsize) + 1, endpoint=True)
        for i, w in enumerate(w_steps):
            # use round to avoid case of floating point limitations in Python
            # the limitation: 1- 0.9 = 0.09999999999998 rather than 0.1
            other_ws_range = round((1 - w), 10)
            w2 = np.linspace(
                0,
                other_ws_range,
                num=round(other_ws_range / stepsize + 1),
                endpoint=True,
            )
            w3 = other_ws_range - w2
            num = w2.shape[0]
            w1 = np.array([w] * num)
            ws = np.hstack(
                [w1.reshape([num, 1]), w2.reshape([num, 1]), w3.reshape([num, 1])]
            )
            if i == 0:
                ws_pairs = ws
            else:
                ws_pairs = np.vstack([ws_pairs, ws])
    else:
        raise Exception(f"{n_objectives} objectives are not supported.")

    assert all(np.round(np.sum(ws_pairs, axis=1), 10) == 1)
    return np.array(ws_pairs)


def plot_po(po: np.ndarray, n_obj: int = 2, title: str = "pf_ap") -> None:
    """Plot pareto-optimal solutions"""
    # po: ndarray (n_solutions * n_objs)
    ## for 2d
    if n_obj == 2:
        po_obj1 = po[:, 0]
        po_obj2 = po[:, 1]

        fig, ax = plt.subplots()
        ax.scatter(po_obj1, po_obj2, marker="o", color="blue")
        ax.plot(po_obj1, po_obj2, color="blue")

        ax.set_xlabel("Obj 1")
        ax.set_ylabel("Obj 2")

        ax.set_title(title)

    elif n_obj == 3:
        po_obj1 = po[:, 0]
        po_obj2 = po[:, 1]
        po_obj3 = po[:, 2]

        plt.figure()
        ax = plt.axes(projection="3d")

        # ax.plot_trisurf(po_obj1, po_obj2, po_obj3, antialiased=True)
        # ax.plot_surface(po_obj1, po_obj2, po_obj3)
        ax.scatter3D(po_obj1, po_obj2, po_obj3, color="blue")

        ax.set_xlabel("Obj 1")
        ax.set_ylabel("Obj 2")
        ax.set_zlabel("Obj 3")

    else:
        raise Exception(
            f"{n_obj} objectives are not supported in the code repository for now!"
        )

    plt.tight_layout()
    plt.show()


def get_default_device() -> th.device:
    return th.device("cuda") if th.cuda.is_available() else th.device("cpu")



from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

from ..concepts import SOProblem


class SOSolver(ABC):
    @abstractmethod
    def solve(
        self,
        problem: SOProblem,
        seed: Optional[int] = None,
    ) -> Tuple[float, Dict[str, float]]:
        """Solve a single-objective optimization problem

        Parameters
        ----------
        problem : SOProblem
            Single-objective optimization problem to solve
        seed : Optional[int], optional
            Random seed, by default None

        Returns
        -------
        Tuple[float, Dict[str, float]]
            A tuple of the objective value and the variables
            that optimize the objective
        """
        ...



from typing import Any

import pytest
import torch as th
from attr import dataclass

from ...utils.interfaces import UdaoEmbedInput, UdaoEmbedItemShape
from ..embedders import BaseEmbedder
from ..model import FixedEmbeddingUdaoModel, UdaoModel
from ..regressors import BaseRegressor


class DummyEmbedder(BaseEmbedder):
    @dataclass
    class Params(BaseEmbedder.Params):
        output_size: int

    @classmethod
    def from_iterator_shape(
        cls, iterator_shape: UdaoEmbedItemShape, **kwargs: Any
    ) -> "DummyEmbedder":
        return cls(cls.Params(**kwargs))

    def forward(self, input: th.Tensor) -> th.Tensor:
        return input


class DummyRegressor(BaseRegressor):
    def forward(self, embedding: th.Tensor, inst_feat: th.Tensor) -> th.Tensor:
        return embedding + inst_feat


@pytest.fixture
def model() -> UdaoModel:
    iterator_shape = UdaoEmbedItemShape(
        embedding_input_shape=1,
        feature_names=["a"],
        output_names=["b"],
    )
    return UdaoModel.from_config(
        regressor_cls=DummyRegressor,
        embedder_cls=DummyEmbedder,
        iterator_shape=iterator_shape,
        regressor_params={},
        embedder_params={"output_size": 1},
    )


@pytest.fixture
def obj_model(model: UdaoModel) -> FixedEmbeddingUdaoModel:
    return FixedEmbeddingUdaoModel(model, None)


class TestUdaoModel:
    def test_from_config(self, model: UdaoModel) -> None:
        assert model.regressor.input_dim == 2
        assert model.embedder.embedding_size == 1
        assert model.regressor.output_dim == 1

    def test_forward(self, model: UdaoModel) -> None:
        embedding = th.tensor([1.0])
        inst_feat = th.tensor([2.0])
        out = model.forward(
            UdaoEmbedInput(embedding_input=embedding, features=inst_feat)
        )
        assert out == th.tensor([3.0])


class TestFixedEmbeddingUdaoModel:
    def test_init(self, obj_model: FixedEmbeddingUdaoModel) -> None:
        assert obj_model.model.regressor.input_dim == 2
        assert obj_model.model.embedder.embedding_size == 1
        assert obj_model.model.regressor.output_dim == 1
        assert obj_model.embedding is None

    def test_forward(self, obj_model: FixedEmbeddingUdaoModel, mocker: Any) -> None:
        embedding = th.tensor([1.0])
        inst_feat = th.tensor([2.0])

        mocked_embedder = mocker.spy(obj_model.model.embedder, "forward")
        mocked_regressor = mocker.spy(obj_model.model.regressor, "forward")
        out1 = obj_model.forward(
            UdaoEmbedInput(embedding_input=embedding, features=inst_feat)
        )
        assert out1 == th.tensor([3.0])
        assert obj_model.embedding == embedding
        out2 = obj_model.forward(
            UdaoEmbedInput(embedding_input=embedding, features=inst_feat)
        )
        assert out2 == th.tensor([3.0])
        assert mocked_embedder.call_count == 1
        assert mocked_regressor.call_count == 2



from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import torch as th
from pandas import DataFrame

from ..containers import BaseContainer, TabularContainer
from ..extractors import FeatureExtractor, TabularFeatureExtractor
from ..iterators import BaseIterator
from ..preprocessors.base_preprocessor import FeaturePreprocessor
from ..utils.utils import DatasetType


@dataclass
class FeaturePipeline:
    extractor: FeatureExtractor
    """Tuple defining the feature extractor and its initialization arguments."""
    preprocessors: Optional[List[FeaturePreprocessor]] = None
    """List of tuples defining feature preprocessors
    and their initialization arguments."""


IT = TypeVar("IT", bound=BaseIterator)


class DataProcessor(Generic[IT]):
    """
    Parameters
    ----------
    iterator_cls: Type[BaseDatasetIterator]
        Dataset iterator class type.

    feature_extractors: Mapping[str, Tuple[FeatureExtractorType, Any]]
        Dict that links a feature name to tuples of the form (Extractor, args)
        where Extractor implements FeatureExtractor and args are the arguments
        to be passed at initialization.
        N.B.: Feature names must match the iterator's parameters.

        If Extractor is a StaticExtractor, the features are extracted
        independently of the split.

        If Extractor is a TrainedExtractor, the extractor is first fitted
        on the train split and then applied to the other splits.

    feature_preprocessors: Optional[Mapping[str, List[FeaturePreprocessor]]]
        Dict that links a feature name to a list of tuples of the form (Processor, args)
        where Processor implements FeatureProcessor and args are the arguments
        to be passed at initialization.
        This allows to apply a series of processors to different features, e.g.
        to normalize the features.
        N.B.: Feature names must match the iterator's parameters.
        If Processor is a StaticExtractor, the features are processed
        independently of the split.

        If Extractor is a TrainedExtractor, the processor is first fitted
        on the train split and then applied to the other splits
        (typically for normalization).

    tensors_dtype: Optional[th.dtype]
        Data type of the tensors returned by the iterator, by default None
    """

    def __init__(
        self,
        iterator_cls: Type[IT],
        feature_extractors: Dict[str, FeatureExtractor],
        feature_preprocessors: Optional[
            Mapping[
                str,
                Sequence[FeaturePreprocessor],
            ]
        ] = None,
        tensors_dtype: Optional[th.dtype] = None,
    ) -> None:
        self.iterator_cls = iterator_cls
        self.feature_extractors = feature_extractors
        self.feature_processors = feature_preprocessors or {}

    def _apply_processing_function(
        self,
        function: Callable[..., BaseContainer],
        data: Union[DataFrame, BaseContainer],
        split: DatasetType,
        is_trained: bool,
    ) -> BaseContainer:
        if is_trained:
            features = function(data, split=split)
        else:
            features = function(data)

        return features

    def extract_features(
        self, data: DataFrame, split: DatasetType
    ) -> Dict[str, BaseContainer]:
        """Extract features for the different splits of the data.

        Returns
        -------
        DataHandler
            self

        Raises
        ------
        ValueError
            Expects data to be split before extracting features.
        """
        features: Dict[str, BaseContainer] = {}
        for name, extractor in self.feature_extractors.items():
            features[name] = self._apply_processing_function(
                extractor.extract_features,
                data,
                split=split,
                is_trained=extractor.trained,
            )
            for preprocessor in self.feature_processors.get(name, []):
                features[name] = self._apply_processing_function(
                    preprocessor.preprocess,
                    features[name],
                    split=split,
                    is_trained=preprocessor.trained,
                )

        return features

    def make_iterator(self, data: DataFrame, keys: Sequence, split: DatasetType) -> IT:
        return self.iterator_cls(keys, **self.extract_features(data, split=split))

    def inverse_transform(
        self, container: TabularContainer, pipeline_name: str
    ) -> DataFrame:
        """Inverse transform the data to the original format.

        Parameters
        ----------
        container: TabularContainer
            Data to be inverse transformed.
        pipeline_name: str
            Name of the feature pipeline to be inverse transformed.
        Returns
        -------
        DataFrame
            Inverse transformed data.
        """

        extractor = self.feature_extractors[pipeline_name]
        if not isinstance(extractor, TabularFeatureExtractor):
            raise ValueError(
                "Only TabularFeatureExtractor supports"
                "transforming back to original dataframe."
            )
        preprocessors = self.feature_processors.get(pipeline_name, [])

        for preprocessor in preprocessors[::-1]:
            if not hasattr(preprocessor, "inverse_transform"):
                raise ValueError(
                    f"Feature preprocessor {pipeline_name} does "
                    "not have an inverse transform method."
                )
            container = preprocessor.inverse_transform(container)  # type: ignore
        df = cast(TabularContainer, container).data
        return df


def create_data_processor(
    iterator_cls: Type[IT], *args: str
) -> Callable[..., DataProcessor[IT]]:
    """
    Creates a function dynamically to instatiate DataProcessor
    based on provided iterator class and additional arguments.

    Parameters
    ----------
    iterator_cls : Type[BaseDatasetIterator]
        Dataset iterator class type.
    args : str
        Additional feature names to be included.

    Returns
    -------
    create_data_processor: Callable[..., DataProcessor]
        A dynamically generated function
        with arguments derived from the provided iterator class,
        in addition to other specified arguments.

    Notes
    -----
    The returned function has the following signature:
        >>> def get_processor(
        >>>    **kwargs: FeaturePipeline,
        >>> ) -> DataProcessor:

        where kwargs are the feature names and their corresponding feature
    """
    params = iterator_cls.get_parameter_names()
    if args is not None:
        params += args

    def get_processor(
        tensors_dtype: Optional[th.dtype] = None,
        **kwargs: FeaturePipeline,
    ) -> DataProcessor[IT]:
        feature_extractors: Dict[str, FeatureExtractor] = {}
        feature_preprocessors: Dict[str, Sequence[FeaturePreprocessor]] = {}
        for param in params:
            if param in kwargs:
                feature_extractors[param] = kwargs[param].extractor
                preprocessors = kwargs[param].preprocessors
                if preprocessors is not None:
                    feature_preprocessors[param] = preprocessors
            else:
                raise ValueError(
                    f"Feature pipeline for {param} not specified in kwargs. "
                    f"All iterator features should be provided with an extractor."
                )

        return DataProcessor[IT](
            iterator_cls=iterator_cls,
            tensors_dtype=tensors_dtype,
            feature_extractors=feature_extractors,
            feature_preprocessors=feature_preprocessors,
        )

    return get_processor



from typing import cast

import numpy as np
import pytest
import torch as th

from ....model.utils.utils import set_deterministic_torch
from ....utils.interfaces import VarTypes
from ...concepts.problem import MOProblem
from ...moo.progressive_frontier import SequentialProgressiveFrontier
from ...soo.mogd import MOGD
from ...utils.moo_utils import Point


@pytest.fixture
def spf(mogd: MOGD) -> SequentialProgressiveFrontier:
    spf = SequentialProgressiveFrontier(
        params=SequentialProgressiveFrontier.Params(),
        solver=mogd,
    )
    return spf


class TestProgressiveFrontier:
    def test__get_corner_points(self, spf: SequentialProgressiveFrontier) -> None:
        utopia = Point(np.array([1, 0.3]))
        nadir = Point(np.array([5, 10]))
        corner_points = spf._get_corner_points(utopia, nadir)
        # 1-------3#
        #         #
        # 0-------2#
        expected_points = [
            Point(np.array([1.0, 0.3])),
            Point(np.array([1.0, 10.0])),
            Point(np.array([5.0, 0.3])),
            Point(np.array([5.0, 10.0])),
        ]
        assert all(c == e for c, e in zip(corner_points, expected_points))

    def test__generate_sub_rectangles_bad(
        self, spf: SequentialProgressiveFrontier
    ) -> None:
        utopia = Point(np.array([1, 0.3]))
        nadir = Point(np.array([5, 10]))
        middle = Point((utopia.objs + nadir.objs) / 2)

        rectangles = spf.generate_sub_rectangles(
            utopia, nadir, middle, successful=False
        )
        ############
        #  0 |  1  #
        ############
        #  - |  -  #
        ############
        assert len(rectangles) == 2
        assert rectangles[0].utopia == Point(np.array([1.0, 5.15]))
        assert rectangles[0].nadir == Point(np.array([3.0, 10]))
        assert rectangles[1].utopia == Point(np.array([3.0, 5.15]))
        assert rectangles[1].nadir == Point(np.array([5.0, 10]))

    def test__generate_sub_rectangles_good(
        self, spf: SequentialProgressiveFrontier
    ) -> None:
        utopia = Point(np.array([1, 0.3]))
        nadir = Point(np.array([5, 10]))
        middle = Point((utopia.objs + nadir.objs) / 2)

        rectangles = spf.generate_sub_rectangles(utopia, nadir, middle)
        ############
        #  1 |  _  #
        ############
        #  0 |  2  #
        ############
        assert len(rectangles) == 3
        assert rectangles[0].utopia == Point(np.array([1.0, 0.3]))
        assert rectangles[0].nadir == Point(np.array([3.0, 5.15]))
        assert rectangles[1].utopia == Point(np.array([1.0, 5.15]))
        assert rectangles[1].nadir == Point(np.array([3.0, 10.0]))
        assert rectangles[2].utopia == Point(np.array([3.0, 0.3]))
        assert rectangles[2].nadir == Point(np.array([5.0, 5.15]))

    def test_get_utopia_and_nadir(self, spf: SequentialProgressiveFrontier) -> None:
        points = [
            Point(np.array([1, 5]), {"v1": 0.2, "v2": 1}),
            Point(np.array([3, 10]), {"v1": 0.8, "v2": 6}),
            Point(np.array([5, 0.3]), {"v1": 0.5, "v2": 3}),
        ]
        utopia, nadir = spf.get_utopia_and_nadir(points)
        np.testing.assert_array_equal(utopia.objs, np.array([1, 0.3]))
        np.testing.assert_array_equal(nadir.objs, np.array([5, 10]))

    def test_solve(
        self,
        spf: SequentialProgressiveFrontier,
        two_obj_problem: MOProblem,
    ) -> None:
        cast(MOGD, spf.solver).patience = 100
        objectives, variables = spf.solve(
            problem=two_obj_problem,
            seed=0,
        )
        assert objectives is not None
        np.testing.assert_array_equal(objectives, np.array([[0, 0]]))
        assert variables[0] == {"v1": 0.0, "v2": 1.0}

    def test_solve_gpu(
        self,
        spf: SequentialProgressiveFrontier,
        two_obj_problem: MOProblem,
    ) -> None:
        if not th.cuda.is_available():
            pytest.skip("Skip GPU test")
        cast(MOGD, spf.solver).device = th.device("cuda")
        cast(MOGD, spf.solver).patience = 100
        objectives, variables = spf.solve(
            problem=two_obj_problem,
            seed=0,
        )
        assert objectives is not None
        np.testing.assert_array_equal(objectives, np.array([[0, 0]]))
        assert variables[0] == {"v1": 0.0, "v2": 1.0}

    def test_solve_non_processed_problem(
        self,
        spf: SequentialProgressiveFrontier,
        simple_problem: MOProblem,
    ) -> None:
        cast(MOGD, spf.solver).patience = 100
        objectives, variables = spf.solve(
            problem=simple_problem,
            seed=0,
        )
        assert objectives is not None
        np.testing.assert_array_almost_equal(objectives, np.array([[1, 1.3]]))
        assert variables[0] == {"v1": 0.0, "v2": 3.0}

    def test_solve_non_processed_problem_gpu(
        self,
        spf: SequentialProgressiveFrontier,
        simple_problem: MOProblem,
    ) -> None:
        if not th.cuda.is_available():
            pytest.skip("Skip GPU test")

        cast(MOGD, spf.solver).device = th.device("cuda")
        cast(MOGD, spf.solver).patience = 100
        objectives, variables = spf.solve(
            problem=simple_problem,
            seed=0,
        )
        assert objectives is not None
        np.testing.assert_array_almost_equal(objectives, np.array([[1, 1.3]]))
        assert variables[0] == {"v1": 0.0, "v2": 3.0}

    def test_get_utopia_and_nadir_raises_when_no_points(
        self, spf: SequentialProgressiveFrontier
    ) -> None:
        with pytest.raises(ValueError):
            spf.get_utopia_and_nadir([])

    def test_get_utopia_and_nadir_raises_when_inconsistent_points(
        self, spf: SequentialProgressiveFrontier
    ) -> None:
        with pytest.raises(Exception):
            spf.get_utopia_and_nadir(
                [
                    Point(np.array([1, 5]), {"v1": 0.2, "v2": 1}),
                    Point(np.array([3, 10]), {"v1": 0.8, "v2": 6}),
                    Point(np.array([5]), {"v1": 0.5, "v2": 3}),
                ]
            )

    def test_get_anchor_points(
        self,
        spf: SequentialProgressiveFrontier,
        two_obj_problem: MOProblem,
    ) -> None:
        set_deterministic_torch()
        anchor_point = spf.get_anchor_point(
            problem=two_obj_problem,
            obj_ind=0,
            seed=0,
        )
        np.testing.assert_array_almost_equal(
            anchor_point.objs, np.array([0.0, 0.6944444])
        )
        assert anchor_point.vars == {"v1": 0.0, "v2": 6.0}
        anchor_point = spf.get_anchor_point(
            problem=two_obj_problem,
            obj_ind=1,
            seed=0,
        )
        np.testing.assert_array_almost_equal(
            anchor_point.objs, np.array([0.301196, 0.0])
        )
        assert anchor_point.vars is not None
        assert anchor_point.vars == {"v1": 0.54881352186203, "v2": 1.0}

    def test_get_anchor_points_with_int(
        self,
        spf: SequentialProgressiveFrontier,
        two_obj_problem: MOProblem,
    ) -> None:
        two_obj_problem.objectives[0].type = VarTypes.INT
        set_deterministic_torch()
        anchor_point = spf.get_anchor_point(problem=two_obj_problem, obj_ind=0, seed=0)
        np.testing.assert_array_almost_equal(anchor_point.objs, np.array([0.0, 0.0]))
        assert anchor_point.vars == {"v1": 0.0, "v2": 1.0}



from abc import ABC, abstractmethod
from typing import Dict, Mapping, Optional, Tuple

import numpy as np
import torch as th

from ..concepts import SOProblem, Variable
from ..utils.exceptions import NoSolutionError
from .so_solver import SOSolver


class SamplerSolver(SOSolver, ABC):
    def __init__(
        self,
        device: Optional[th.device] = None,
    ) -> None:
        super().__init__()

        self.device = device or th.device("cpu")

    @abstractmethod
    def _get_input(
        self, variables: Mapping[str, Variable], seed: Optional[int] = None
    ) -> Dict[str, np.ndarray]:
        """
        generate samples of variables

        Parameters
        -----------
        variables: List[Variable],
            lower and upper var_ranges of variables(non-ENUM),
            and values of ENUM variables
        seed: int, by default None
            random seed

        Returns
        --------
        np.ndarray,
            variables (n_samples * n_vars)
        """
        pass

    def solve(
        self, problem: SOProblem, seed: Optional[int] = None
    ) -> Tuple[float, Dict[str, float]]:
        """Solve a single-objective optimization problem

        Parameters
        ----------
        objective : Objective
            Objective to be optimized
        variables : Dict[str, Variable]
            Variables to be optimized
        constraints : Optional[List[Constraint]], optional
            List of constraints to comply with, by default None
        input_parameters : Optional[Dict[str, Any]], optional
            Fixed parameters input to objectives and/or constraints,
            by default None
        seed: int, by default None
            random seed
        Returns
        -------
        Point
            A point that satisfies the constraints
            and optimizes the objective

        Raises
        ------
        NoSolutionError
            If no feasible solution is found
        """
        if self.device:
            for constraint in problem.constraints:
                constraint.to(self.device)
            problem.objective.to(self.device)

        if problem.constraints is None:
            pass
        variable_values = self._get_input(problem.variables, seed=seed)
        filtered_vars = self.filter_on_constraints(
            input_vars=variable_values,
            problem=problem,
        )
        if any([len(v) == 0 for v in filtered_vars.values()]):
            raise NoSolutionError("No feasible solution found!")

        th_value = problem.apply_function(
            optimization_element=problem.objective,
            input_variables=filtered_vars,
            device=self.device,
        )
        objective_value = th_value.detach().cpu().numpy().reshape(-1, 1)
        op_ind = int(np.argmin(objective_value * problem.objective.direction))

        return (
            objective_value[op_ind],
            {k: v[op_ind] for k, v in filtered_vars.items()},
        )

    def filter_on_constraints(
        self,
        input_vars: Dict[str, np.ndarray],
        problem: SOProblem,
    ) -> Dict[str, np.ndarray]:
        """Keep only input variables that don't violate constraints

        Parameters
        ----------
        wl_id : str | None
            workload id
        input_vars : np.ndarray
            input variables
        constraints : List[Constraint]
            constraint functions
        """
        if not problem.constraints:
            return input_vars

        available_indices = np.arange(len(next(iter(input_vars.values()))))
        for constraint in problem.constraints:
            const_values = (
                problem.apply_function(constraint, input_vars, device=self.device)
                .detach()
                .cpu()
                .numpy()
            )

            if constraint.upper is not None:
                available_indices = np.intersect1d(
                    available_indices, np.where(const_values <= constraint.upper)
                )
            if constraint.lower is not None:
                available_indices = np.intersect1d(
                    available_indices, np.where(const_values >= constraint.lower)
                )
        filtered_vars = {k: v[available_indices] for k, v in input_vars.items()}
        return filtered_vars



from .base_container import BaseContainer
from .query_structure_container import QueryStructureContainer
from .tabular_container import TabularContainer

__all__ = [
    "BaseContainer",
    "TabularContainer",
    "QueryStructureContainer",
]



import pytest
import torch as th

from ...regressors.mlp import MLP
from ...utils.utils import set_deterministic_torch


@pytest.mark.parametrize(
    "embed_dim, feat_dim, expected_output",
    [
        (
            10,
            10,
            th.tensor(
                [
                    [1.175319e00, 2.838397e00, 1.457014e00, 1.331543e00, 6.580783e-01],
                    [
                        1.730856e00,
                        6.496964e-01,
                        3.652042e-01,
                        7.815728e-01,
                        8.088993e-01,
                    ],
                ]
            ),
        ),
        (
            5,
            5,
            th.tensor(
                [
                    [1.338640e00, 2.081545e00, 1.356574e00, 5.648745e-01, 5.369927e-01],
                    [1.742966e00, 8.478099e-01, 1.723262e00, 1.659415e00, 8.338411e-01],
                ]
            ),
        ),
    ],
)
def test_udao_mlp_forward_shape(
    embed_dim: int, feat_dim: int, expected_output: th.Tensor
) -> None:
    set_deterministic_torch(0)
    sample_mlp_params = MLP.Params(
        input_embedding_dim=embed_dim,
        input_features_dim=feat_dim,
        output_dim=5,
        n_layers=3,
        hidden_dim=5,
        dropout=0,
    )
    model = MLP(sample_mlp_params)
    sample_embedding = th.rand((2, embed_dim))
    sample_inst_feat = th.rand((2, feat_dim))
    output = model.forward(sample_embedding, sample_inst_feat)
    assert output.shape == (2, 5)
    assert th.allclose(output, expected_output, rtol=1e-5)



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Type, cast

import lightning.pytorch as pl
import torch as th
from lightning.pytorch.utilities.types import OptimizerLRScheduler
from torch import nn
from torch.nn.modules.loss import _Loss
from torchmetrics import Metric, MetricCollection

from .utils.losses import WMAPELoss


@dataclass
class LearningParams:
    init_lr: float = 1e-3
    """Initial learning rate."""
    weight_decay: float = 1e-2
    """Weight decay."""
    min_lr: float = 1e-5
    """Minimum learning rate."""


class UdaoModule(pl.LightningModule):
    """Pytorch Lightning module for UDAO.

    Parameters
    ----------
    model : nn.Module
        The model to train.
    metrics: Optional[List[Type[Metric]]], optional
        A list of metric classes - from torchmetrics - to compute,
        by default None
    objectives : List[str]
        The list of objectives to train on.
        Should be a subset of the objectives in the dataset, i.e.
        the size of the output of the model should be equal or
        larger than the size of the list of objectives.
    loss : Optional[_Loss], optional
        A ptorch loss function to apply, by default None
        If None, the WMAPELoss is used.
    loss_weights : Optional[Dict[str, float]], optional
        Loss weights to apply in sum of different
        objective losses, by default None
    learning_params : Optional[LearningParams], optional
        Learning parameters, by default None
        Default values are used if None.
    """

    def __init__(
        self,
        model: nn.Module,
        objectives: List[str],
        loss: Optional[_Loss] = None,
        metrics: Optional[List[Type[Metric]]] = None,
        loss_weights: Optional[Dict[str, float]] = None,
        learning_params: Optional[LearningParams] = None,
    ) -> None:
        super().__init__()
        self.model = model
        self.loss: _Loss = WMAPELoss()
        if loss is not None:
            self.loss = loss

        self.objectives = objectives
        self.lr_scheduler: Optional[th.optim.lr_scheduler.LRScheduler] = None
        if learning_params is None:
            self.learning_params = LearningParams()
        else:
            self.learning_params = learning_params
        self.optimizer: th.optim.Optimizer = th.optim.AdamW(
            self.model.parameters(),
            lr=self.learning_params.init_lr,
            weight_decay=self.learning_params.weight_decay,
        )
        self.loss_weight: Dict[str, float]
        if loss_weights is None:
            self.loss_weights = {m: 1.0 for m in self.objectives}
        else:
            self.loss_weights = loss_weights

        if metrics:
            self.metrics = nn.ModuleDict(
                {
                    objective: MetricCollection(
                        [m() for m in metrics], prefix=f"{objective}_"
                    )
                    for objective in self.objectives
                }
            )

    def compute_loss(
        self,
        y: th.Tensor,
        y_hat: th.Tensor,
    ) -> Tuple[th.Tensor, Dict[str, th.Tensor]]:
        """Compute the loss for different objectives
        and sum them with given weights, if more than one."""
        loss_dict: Dict[str, th.Tensor] = {
            m: self.loss(y_hat[:, i], y[:, i]) for i, m in enumerate(self.objectives)
        }

        loss = th.sum(
            th.stack([self.loss_weights[k] * loss for k, loss in loss_dict.items()])
        )
        return loss, loss_dict

    def configure_optimizers(self) -> OptimizerLRScheduler:
        return self.optimizer

    def _shared_step(
        self, batch: Tuple[Any, th.Tensor], split: str
    ) -> Tuple[th.Tensor, th.Tensor]:
        """Update metrics for a given split."""
        features, y = batch
        y_hat = self.model(features)
        for i, objective in enumerate(self.objectives):
            cast(Metric, self.metrics[objective]).update(y_hat[:, i], y[:, i])
        return y_hat, y

    def _shared_epoch_end(self, split: str) -> None:
        """Compute and log metrics of all objectives for a given split."""
        for objective in self.objectives:
            metric = cast(Metric, self.metrics[objective])
            output = metric.compute()
            for k, v in output.items():
                self.log(f"{split}_{k}", v, on_epoch=True, prog_bar=True, logger=True)
            metric.reset()

    def training_step(self, batch: Tuple[Any, th.Tensor], batch_idx: int) -> th.Tensor:
        y_hat, y = self._shared_step(batch, "train")
        loss, _ = self.compute_loss(y, y_hat)
        if th.isnan(loss):
            raise ValueError("got a nan loss in train")
        elif th.isinf(loss):
            raise ValueError("got an inf loss in train")
        self.log(
            "train_loss", loss, on_step=True, on_epoch=True, prog_bar=True, logger=True
        )
        return loss

    def on_train_epoch_end(self) -> None:
        self._shared_epoch_end("train")

    def validation_step(self, batch: Tuple[Any, th.Tensor], batch_idx: int) -> None:
        self._shared_step(batch, "val")

    def on_validation_epoch_end(self) -> None:
        self._shared_epoch_end("val")

    def test_step(self, batch: Tuple[Any, th.Tensor], batch_idx: int) -> None:
        self._shared_step(batch, "test")

    def on_test_epoch_end(self) -> None:
        self._shared_epoch_end("test")



import torch
import torch.nn as nn


class IsoBN(nn.Module):
    """Isotropic Batch Normalization."""

    def __init__(self, hidden_dim: int) -> None:
        super().__init__()
        self.cov = nn.Parameter(torch.zeros(hidden_dim, hidden_dim))
        self.std = nn.Parameter(torch.zeros(hidden_dim))

    def forward(
        self,
        input: torch.Tensor,
        momentum: float = 0.05,
        eps: float = 1e-3,
        beta: float = 0.5,
    ) -> torch.Tensor:
        with torch.no_grad():
            if self.training:
                x = input.detach()
                n = x.size(0)
                mean = x.mean(dim=0)
                y = x - mean.unsqueeze(0)
                std = (y**2).mean(0) ** 0.5
                cov = (y.t() @ y) / n
                self.cov.data += momentum * (cov.data - self.cov.data)
                self.std.data += momentum * (std.data - self.std.data)

            corr = torch.clamp(
                self.cov
                / (
                    torch.ger(self.std, self.std)
                    + torch.full_like(torch.ger(self.std, self.std), 1e-6)
                ),
                -1,
                1,
            )
            gamma = (corr**2).mean(1)
            denorm = gamma * self.std
            scale = 1 / (denorm + eps) ** beta
            E = torch.diag(self.cov).sum()
            new_E = (torch.diag(self.cov) * (scale**2)).sum()
            m = (E / (new_E + eps)) ** 0.5
            scale *= m
        return input * scale.unsqueeze(0).detach()



from typing import Dict, Optional, Type

import torch as th
from torch import nn

from ..utils.interfaces import UdaoEmbedInput, UdaoEmbedItemShape
from .embedders.base_embedder import BaseEmbedder
from .regressors.base_regressor import BaseRegressor


class UdaoModel(nn.Module):
    @classmethod
    def from_config(
        cls,
        embedder_cls: Type[BaseEmbedder],
        regressor_cls: Type[BaseRegressor],
        iterator_shape: UdaoEmbedItemShape,
        regressor_params: Dict,
        embedder_params: Dict,
    ) -> "UdaoModel":
        embedder = embedder_cls.from_iterator_shape(iterator_shape, **embedder_params)
        regressor = regressor_cls(
            regressor_cls.Params(
                input_embedding_dim=embedder.embedding_size,
                input_features_dim=len(iterator_shape.feature_names),
                output_dim=len(iterator_shape.output_names),
                **regressor_params,
            ),
        )
        return cls(embedder, regressor)
        pass

    def __init__(self, embedder: BaseEmbedder, regressor: BaseRegressor) -> None:
        super().__init__()
        self.embedder = embedder
        self.regressor = regressor

    def forward(
        self, input_data: UdaoEmbedInput, embedding: Optional[th.Tensor] = None
    ) -> th.Tensor:
        if embedding is None:
            embedding = self.embedder(input_data.embedding_input)
        inst_feat = input_data.features
        return self.regressor(embedding, inst_feat)


class FixedEmbeddingUdaoModel(th.nn.Module):
    """
    Assumes the embedding part does not change between inputs and caches
    it after computing it once. This is relevant for an optimization pipeline
    where the embedding input (e.g. a query plan) is a fixed parameter
    """

    def __init__(self, model: UdaoModel, embedding: Optional[th.Tensor] = None) -> None:
        super().__init__()
        self.model = model
        self.embedding = embedding

    def forward(self, input_data: UdaoEmbedInput) -> th.Tensor:
        if self.embedding is None:
            self.embedding = self.model.embedder(input_data.embedding_input)
        return self.model(input_data, self.embedding)



import json
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch as th

from ..concepts import Objective
from ..concepts.problem import MOProblem
from ..soo.mogd import MOGD
from ..soo.so_solver import SOSolver
from ..utils import moo_utils as moo_ut
from ..utils.exceptions import NoSolutionError
from ..utils.moo_utils import Point, get_default_device
from .mo_solver import MOSolver


class WeightedSumObjective(Objective):
    """Weighted Sum Objective"""

    def __init__(
        self,
        problem: MOProblem,
        ws: List[float],
        allow_cache: bool = False,
        normalize: bool = True,
        device: Optional[th.device] = None,
    ) -> None:
        self.device = device or get_default_device()
        self.problem = problem
        self.ws = ws
        super().__init__(name="weighted_sum", function=self.function, minimize=True)
        self._cache: Dict[str, th.Tensor] = {}
        self.allow_cache = allow_cache
        self.normalize = normalize

    def _function(self, *args: Any, **kwargs: Any) -> th.Tensor:
        hash_var = ""
        if self.allow_cache:
            hash_var = json.dumps(str(args) + str(kwargs))
            if hash_var in self._cache:
                return self._cache[hash_var]
        objs: List[th.Tensor] = []
        for objective in self.problem.objectives:
            obj = objective(*args, **kwargs) * objective.direction
            objs.append(obj.squeeze())
        objs_tensor = th.vstack(objs).T
        # shape (n_feasible_samples/grids, n_objs)
        if self.allow_cache:
            self._cache[hash_var] = objs_tensor
        return objs_tensor

    def function(self, *args: Any, **kwargs: Any) -> th.Tensor:
        """Sum of weighted normalized objectives"""
        objs_tensor = self._function(*args, **kwargs)
        if self.normalize:
            objs_tensor = self._normalize_objective(objs_tensor)
        return th.sum(objs_tensor * th.tensor(self.ws, device=self.device), dim=1)

    def _normalize_objective(self, objs_array: th.Tensor) -> th.Tensor:
        """Normalize objective values to [0, 1]

        Parameters
        ----------
        objs_array : np.ndarray
            shape (n_feasible_samples/grids, n_objs)

        Returns
        -------
        np.ndarray
            shape (n_feasible_samples/grids, n_objs)

        Raises
        ------
        NoSolutionError
            if lower bounds of objective values are
            higher than their upper bounds
        """
        objs_min, objs_max = th.min(objs_array, 0).values, th.max(objs_array, 0).values
        if th.any((objs_min - objs_max) > 0):
            raise NoSolutionError(
                "Cannot do normalization! Lower bounds of "
                "objective values are higher than their upper bounds."
            )
        elif th.all((objs_min - objs_max) == 0):
            return th.zeros_like(objs_array)
        return (objs_array - objs_min) / (objs_max - objs_min)

    def to(self, device: Optional[th.device] = None) -> "WeightedSumObjective":
        """Move objective to device"""
        if device is None:
            device = get_default_device()

        self.device = device
        for objective in self.problem.objectives:
            objective.to(device)
        for constraint in self.problem.constraints:
            constraint.to(device)
        self._cache = {k: v.to(device) for k, v in self._cache.items()}
        return self


class WeightedSum(MOSolver):
    """
    Weighted Sum (WS) algorithm for MOO

    Parameters
    ----------
    ws_pairs: np.ndarray,
        weight settings for all objectives, of shape (n_weights, n_objs)
    inner_solver: BaseSolver,
        the solver used in Weighted Sum
    objectives: List[Objective],
        objective functions
    constraints: List[Constraint],
        constraint functions

    """

    @dataclass
    class Params:
        ws_pairs: np.ndarray
        """weight sets for all objectives, of shape (n_weights, n_objs)"""
        so_solver: SOSolver
        """solver for SOO"""
        normalize: bool = True
        """whether to normalize objective values to [0, 1] before applying WS"""
        allow_cache: bool = False
        """whether to cache the objective values"""
        device: Optional[th.device] = field(default_factory=get_default_device)
        """device on which to perform torch operations, by default available device."""

    def __init__(
        self,
        params: Params,
    ):
        super().__init__()
        self.so_solver = params.so_solver
        self.ws_pairs = params.ws_pairs
        self.allow_cache = params.allow_cache
        self.normalize = params.normalize
        self.device = params.device

        if self.allow_cache and isinstance(params.so_solver, MOGD):
            raise NotImplementedError(
                "MOGD does not support caching." "Please set allow_cache=False."
            )

    def solve(
        self, problem: MOProblem, seed: Optional[int] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """solve MOO problem by Weighted Sum (WS)

        Parameters
        ----------
        variables : List[Variable]
            List of the variables to be optimized.
        input_parameters : Optional[Dict[str, Any]]
            Fixed input parameters expected by
            the objective functions.

        Returns
        -------
        Tuple[Optional[np.ndarray],Optional[np.ndarray]]
            Pareto solutions and corresponding variables.
        """
        candidate_points: List[Point] = []
        objective = WeightedSumObjective(
            problem, self.ws_pairs[0], self.allow_cache, self.normalize, self.device
        )
        so_problem = problem.derive_SO_problem(objective)
        for i, ws in enumerate(self.ws_pairs):
            objective.ws = ws
            _, soo_vars = self.so_solver.solve(
                so_problem,
                seed=seed + i * (not self.allow_cache) if seed is not None else None,
            )

            objective_values = np.array(
                [
                    problem.apply_function(obj, soo_vars, device=self.device)
                    .cpu()
                    .numpy()
                    for obj in problem.objectives
                ]
            ).T.squeeze()

            candidate_points.append(Point(objective_values, soo_vars))

        return moo_ut.summarize_ret(
            [point.objs for point in candidate_points],
            [point.vars for point in candidate_points],
        )



import pandas as pd
import torch as th

from ...containers.tabular_container import TabularContainer
from ...iterators.tabular_iterator import TabularIterator


def test_iterator_augmentation() -> None:
    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "id": ["1", "2", "3"]})
    df.set_index("id", inplace=True)
    tabular_container = TabularContainer(data=df)
    iterator = TabularIterator(["1", "2", "3"], tabular_container)
    assert th.equal(iterator[0], th.tensor([1, 4]))
    iterator.set_augmentations([lambda x: x + 1, lambda x: x**2])
    assert th.equal(iterator[0], th.tensor([(1 + 1) ** 2, (4 + 1) ** 2]))



from typing import cast

import numpy as np
import pandas as pd
import pytest
from sklearn.preprocessing import MinMaxScaler

from ....utils.interfaces import VarTypes
from ...containers.tabular_container import TabularContainer
from ...extractors.tabular_extractor import TabularFeatureExtractor
from ...handler.data_processor import (
    DataProcessor,
    FeaturePipeline,
    create_data_processor,
)
from ...iterators.tabular_iterator import TabularIterator
from ...preprocessors.normalize_preprocessor import NormalizePreprocessor


@pytest.fixture
def sample_df() -> pd.DataFrame:
    n = 1000
    ids = list(range(1, n + 1))
    tids = [1 for _ in range(n - 10)] + [2 for _ in range(10)]
    values = list(range(n))

    df = pd.DataFrame.from_dict({"id": ids, "tid": tids, "value": values})
    df.set_index("id", inplace=True)

    return df


@pytest.fixture
def data_processor() -> DataProcessor:
    data_processor = DataProcessor(
        iterator_cls=TabularIterator,
        feature_extractors={
            "tabular_feature": TabularFeatureExtractor(
                columns={"value": VarTypes.FLOAT}
            )
        },
        feature_preprocessors={
            "tabular_feature": [NormalizePreprocessor(MinMaxScaler())]
        },
    )

    return data_processor


def test_create_data_processor() -> None:
    # Create the dynamic getter
    data_processor_getter = create_data_processor(TabularIterator)

    scaler = MinMaxScaler()
    # Instantiate the dynamic class
    params_instance = data_processor_getter(
        tabular_feature=FeaturePipeline(
            extractor=TabularFeatureExtractor(columns=["col1", "col2"]),
            preprocessors=[NormalizePreprocessor(scaler)],
        ),
    )

    # Test if the provided parameters exist and are set correctly
    assert params_instance.iterator_cls == TabularIterator
    assert isinstance(
        params_instance.feature_extractors["tabular_feature"], TabularFeatureExtractor
    )

    assert params_instance.feature_processors is not None
    assert len(params_instance.feature_processors) == 1
    assert isinstance(
        params_instance.feature_processors["tabular_feature"][0], NormalizePreprocessor
    )


class TestDataProcessor:
    def test_extract_features_applies_normalization(
        self, sample_df: pd.DataFrame, data_processor: DataProcessor
    ) -> None:
        features = data_processor.extract_features(sample_df, "train")

        assert set(features.keys()) == {"tabular_feature"}
        assert cast(TabularContainer, features["tabular_feature"]).data.shape == (
            len(sample_df),
            1,
        )
        np.testing.assert_array_almost_equal(
            cast(TabularContainer, features["tabular_feature"]).data.values,
            np.linspace(0, 1, len(sample_df)).reshape(-1, 1),
        )

    def test_make_iterators(
        self, sample_df: pd.DataFrame, data_processor: DataProcessor
    ) -> None:
        iterator = data_processor.make_iterator(
            keys=list(sample_df.index), data=sample_df, split="train"
        )
        assert iterator[0] == 0

    def test_inverse_transform(
        self, sample_df: pd.DataFrame, data_processor: DataProcessor
    ) -> None:
        features = data_processor.extract_features(sample_df, "train")
        assert isinstance(features["tabular_feature"], TabularContainer)
        df_inverse = data_processor.inverse_transform(
            container=features["tabular_feature"], pipeline_name="tabular_feature"
        )
        np.testing.assert_array_almost_equal(
            df_inverse[["value"]].values, sample_df[["value"]].values
        )



import torch as th

from ...utils.metrics import ErrorPercentiles, QErrorPercentiles


class TestErrorPercentile:
    def test_compute(self) -> None:
        input_tensor = th.tensor([1, 2.2, 3.3])
        target_tensor = th.tensor([1, 2, 3])
        metric = ErrorPercentiles(percentiles=[50])
        metric.update(input_tensor, target_tensor)
        result = metric.compute()
        assert th.allclose(result["mean"], th.tensor(0.2 / 3))
        assert th.allclose(result["q_50"], th.tensor(0.1))
        metric.reset()
        assert metric.error_rate.shape == (0,)


class TestQErrorPercentile:
    def test_compute(self) -> None:
        input_tensor = th.tensor([1, 2.2, 3.3])
        target_tensor = th.tensor([1, 2, 3])
        metric = QErrorPercentiles(percentiles=[50])
        metric.update(input_tensor, target_tensor)
        result = metric.compute()
        assert th.allclose(result["mean"], th.tensor(1 + 0.2 / 3))
        assert th.allclose(result["q_50"], th.tensor(1.1))
        metric.reset()
        assert metric.q_error.shape == (0,)



from typing import Dict, Optional

import pandas as pd
import pytest

from ....utils.interfaces import VarTypes
from ...containers import TabularContainer
from ...extractors import TabularFeatureExtractor


class TestTabularFeatureExtractor:
    def test_extract_features_with_list(self) -> None:
        # Create a sample DataFrame
        df = pd.DataFrame(
            {
                "int_col": [1, 2, 3],
                "float_col": [1.1, 2.2, 3.3],
                "bool_col": [True, False, True],
                "enum_col": ["a", "b", "a"],
            }
        )
        columns = ["int_col", "float_col"]
        extractor = TabularFeatureExtractor(columns=columns)
        result_container = extractor.extract_features(df)

        # Check if the result is a TabularContainer
        assert isinstance(result_container, TabularContainer)
        # Check if the extracted features contain only the specified columns
        assert list(result_container.data.columns) == columns

    def test_extract_features_with_dict(self) -> None:
        df = pd.DataFrame(
            {
                "int_col": [1, 2, 3],
                "float_col": [1.1, 2.2, 3.3],
                "bool_col": [True, False, True],
                "enum_col": ["a", "b", "a"],
            }
        )
        columns: Dict[str, Optional[VarTypes]] = {
            "int_col": VarTypes.INT,
            "float_col": VarTypes.FLOAT,
            "bool_col": VarTypes.BOOL,
            "enum_col": VarTypes.CATEGORY,
        }
        extractor = TabularFeatureExtractor(columns=columns)
        result_container = extractor.extract_features(df)

        # Check if the result is a TabularContainer
        assert isinstance(result_container, TabularContainer)
        # Check if the extracted DataFrame has the right types
        assert result_container.data["int_col"].dtype == "int"
        assert result_container.data["float_col"].dtype == "float"
        assert result_container.data["bool_col"].dtype == "bool"
        assert result_container.data["enum_col"].dtype.name == "category"

    def test_extract_features_with_invalid_var_type(self) -> None:
        df = pd.DataFrame({"int_col": [1, 2, 3]})
        columns: Dict[str, Optional[VarTypes]] = {"int_col": "INVALID_TYPE"}  # type: ignore
        extractor = TabularFeatureExtractor(columns=columns)

        with pytest.raises(Exception) as exc_info:
            extractor.extract_features(df)
        assert "Unknown variable type" in str(exc_info.value)

    def test_extract_features_with_none_var_type(self) -> None:
        df = pd.DataFrame({"int_col": [1, 2, 3], "float_col": [1.1, 2.2, 3.3]})
        columns = {"int_col": None, "float_col": VarTypes.FLOAT}
        extractor = TabularFeatureExtractor(columns=columns)
        result_container = extractor.extract_features(df)

        assert isinstance(result_container, TabularContainer)
        assert "int_col" in result_container.data.columns
        assert result_container.data["float_col"].dtype == "float"



from dataclasses import dataclass
from typing import Dict, List, Literal, Optional

import dgl
import torch as th
import torch.nn as nn

from .base_graph_embedder import BaseGraphEmbedder
from .layers.graph_transformer_layer import GraphTransformerLayer
from .layers.multi_head_attention import AttentionLayerName

ReadoutType = Literal["sum", "max", "mean"]


class GraphTransformer(BaseGraphEmbedder):
    """Graph Transformer Network
    Computes graph embedding using attention mechanism
    (either QF, RAAL, or GTN)
    """

    @dataclass
    class Params(BaseGraphEmbedder.Params):
        pos_encoding_dim: int
        """Dimension of the position encoding."""
        n_layers: int
        """Number of GCN layers."""
        n_heads: int
        """Number of attention heads."""
        hidden_dim: int
        """Size of the hidden layers outputs."""
        readout: ReadoutType
        """Readout type: how the node embeddings are aggregated
        to form the graph embedding."""
        max_dist: Optional[int] = None
        """Maximum distance for QF attention."""
        non_siblings_map: Optional[Dict[int, Dict[int, List[int]]]] = None
        """Non-siblings map for RAAL attention.
        For each type of graph, maps the edge id to
        all nodes that are not siblings of its source node"""
        attention_layer_name: AttentionLayerName = "GTN"
        """Defines which attention layer to use (QF, RAAL, or GTN))"""
        dropout: float = 0.0
        """Dropout probability."""
        residual: bool = True
        """Whether to make the layer residual. Defaults to True."""
        use_bias: bool = False
        """Whether to use bias in the attention layer. Defaults to False."""
        batch_norm: bool = True
        """Whether to use batch normalization. Defaults to True."""
        layer_norm: bool = False
        """Whether to use layer normalization. Defaults to False."""

    def __init__(self, net_params: Params) -> None:
        super().__init__(net_params=net_params)
        self.attention_layer_name = net_params.attention_layer_name
        self.embedding_lap_pos_enc = nn.Linear(
            net_params.pos_encoding_dim, net_params.hidden_dim
        )
        self.embedding_h = nn.Linear(self.input_size, net_params.hidden_dim)
        self.readout = net_params.readout
        self.attention_bias = None
        if self.attention_layer_name == "QF":
            if net_params.max_dist is None:
                raise ValueError("max_dist is required for QF")
            max_dist = net_params.max_dist
            self.attention_bias = nn.Parameter(th.zeros(max_dist))

        elif self.attention_layer_name == "RAAL":
            if net_params.non_siblings_map is None:
                raise ValueError("non_siblings_map is required for RAAL")
        elif self.attention_layer_name == "GTN":
            pass
        else:
            raise ValueError(self.attention_layer_name)

        if net_params.hidden_dim != net_params.output_size:
            raise ValueError(
                f"hidden_dim ({net_params.hidden_dim}) must be "
                f"equal to output_size ({net_params.output_size})"
            )
        self.layers = nn.ModuleList(
            [
                GraphTransformerLayer(
                    in_dim=net_params.hidden_dim,
                    out_dim=out_dim,
                    n_heads=net_params.n_heads,
                    dropout=net_params.dropout,
                    layer_norm=net_params.layer_norm,
                    batch_norm=net_params.batch_norm,
                    residual=net_params.residual,
                    use_bias=net_params.use_bias,
                    attention_layer_name=net_params.attention_layer_name,
                    attention_bias=self.attention_bias,
                    non_siblings_map=net_params.non_siblings_map,
                )
                for out_dim in [
                    net_params.hidden_dim
                    if i < net_params.n_layers - 1
                    else net_params.output_size
                    for i in range(net_params.n_layers)
                ]
            ]
        )

    def _embed(self, g: dgl.DGLGraph, h: th.Tensor) -> th.Tensor:
        h_lap_pos_enc = g.ndata["pos_enc"]
        h = self.embedding_h(h)
        h_lap_pos_enc = self.embedding_lap_pos_enc(h_lap_pos_enc)
        h = h + h_lap_pos_enc

        for layer in self.layers:
            h = layer.forward(g, h)
        g.ndata["h"] = h

        if self.readout == "sum":
            hg = dgl.sum_nodes(g, "h")
        elif self.readout == "max":
            hg = dgl.max_nodes(g, "h")
        elif self.readout == "mean":
            hg = dgl.mean_nodes(g, "h")
        else:
            raise NotImplementedError
        return hg

    def forward(self, g: dgl.DGLGraph) -> th.Tensor:  # type: ignore[override]
        h = self.concatenate_op_features(g)
        return self.normalize_embedding(self._embed(g, h))



from .containers import BaseContainer, QueryStructureContainer, TabularContainer
from .extractors import (
    FeatureExtractor,
    PredicateEmbeddingExtractor,
    QueryStructureExtractor,
    StaticExtractor,
    TabularFeatureExtractor,
    TrainedExtractor,
)
from .handler.data_handler import DataHandler
from .handler.data_processor import DataProcessor
from .iterators import BaseIterator, QueryPlanIterator, TabularIterator, UdaoIterator
from .preprocessors import (
    NormalizePreprocessor,
    OneHotPreprocessor,
    StaticPreprocessor,
    TrainedPreprocessor,
)

__all__ = [
    "DataHandler",
    "DataProcessor",
    "TabularIterator",
    "QueryPlanIterator",
    "UdaoIterator",
    "BaseIterator",
    "TabularFeatureExtractor",
    "QueryStructureExtractor",
    "StaticExtractor",
    "TrainedExtractor",
    "PredicateEmbeddingExtractor",
    "FeatureExtractor",
    "TabularContainer",
    "QueryStructureContainer",
    "BaseContainer",
    "NormalizePreprocessor",
    "OneHotPreprocessor",
    "StaticPreprocessor",
    "TrainedPreprocessor",
]



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from torch import nn

from udao.utils.interfaces import UdaoEmbedItemShape


class BaseEmbedder(nn.Module, ABC):
    """Placeholder Base class for Embedder networks."""

    @dataclass
    class Params:
        output_size: int
        """The size of the output embedding."""

    @classmethod
    @abstractmethod
    def from_iterator_shape(
        cls,
        iterator_shape: UdaoEmbedItemShape,
        **kwargs: Any,
    ) -> "BaseEmbedder":
        ...

    def __init__(self, net_params: Params) -> None:
        super().__init__()
        self.embedding_size = net_params.output_size



from typing import Dict, Literal, Optional

import pandas as pd
from sklearn.model_selection import train_test_split

PandasTypes = {float: "float64", int: "int64", str: "object"}

DatasetType = Literal["train", "val", "test"]


def train_test_val_split_on_column(
    df: pd.DataFrame,
    groupby_col: Optional[str],
    *,
    val_frac: float,
    test_frac: float,
    random_state: Optional[int] = None,
) -> Dict[DatasetType, pd.DataFrame]:
    """return a dictionary of DatasetType (train/val/test) and the DataFrame"""
    train_df, non_train_df = train_test_split(
        df,
        test_size=val_frac + test_frac,
        stratify=df[groupby_col] if groupby_col else None,
        random_state=random_state,
    )
    val_df, test_df = train_test_split(
        non_train_df,
        test_size=test_frac / (val_frac + test_frac),
        stratify=non_train_df[groupby_col] if groupby_col else None,
        random_state=random_state,
    )
    df_dict: Dict[DatasetType, pd.DataFrame] = {
        "train": train_df,
        "val": val_df,
        "test": test_df,
    }
    return df_dict



from dataclasses import dataclass
from typing import Dict, Iterable, Optional

import dgl
import pandas as pd

from ...data.containers.base_container import BaseContainer
from ..utils.query_plan import QueryPlanStructure


@dataclass
class QueryDescription:
    template_graph: dgl.DGLGraph
    graph_features: Iterable
    meta_features: Optional[Iterable]
    operation_types: Iterable


@dataclass
class QueryStructureContainer(BaseContainer):
    """Container for the query structure and features of a query plan."""

    graph_features: pd.DataFrame
    """ Stores the features of the operations in the query plan."""
    graph_meta_features: Optional[pd.DataFrame]
    """ Stores the meta features of the operations in the query plan."""
    template_plans: Dict[int, QueryPlanStructure]
    """Link a template id to a QueryPlanStructure"""
    key_to_template: Dict[str, int]
    """Link a key to a template id."""
    operation_types: pd.Series
    """Stores the operation types of the operations in the query plan."""

    def get(self, key: str) -> QueryDescription:
        graph_features = self.graph_features.loc[key].values
        graph_meta_features = (
            None
            if self.graph_meta_features is None
            else self.graph_meta_features.loc[key].values
        )
        template_id = self.key_to_template[key]
        template_graph = self.template_plans[template_id].graph.clone()
        operation_types = self.operation_types.loc[key].values
        return QueryDescription(
            template_graph,
            graph_features,
            graph_meta_features,
            operation_types,
        )



import torch as th
import torch.nn.functional as F
from torch.nn.modules.loss import _Loss

from .utils import EPS


class WMAPELoss(_Loss):
    def __init__(self) -> None:
        super().__init__(reduction="sum")

    def forward(self, input: th.Tensor, target: th.Tensor) -> th.Tensor:
        return th.abs(input - target).sum() / th.clip(target.sum(), EPS)


class MSLELoss(_Loss):
    def forward(self, input: th.Tensor, target: th.Tensor) -> th.Tensor:
        return F.mse_loss(
            th.log(th.clip(input, min=EPS)),
            th.log(th.clip(target, min=EPS)),
            reduction=self.reduction,
        )


class MAPELoss(_Loss):
    def forward(self, input: th.Tensor, target: th.Tensor) -> th.Tensor:
        return th.mean(th.abs(input - target) / (th.clip(target, EPS)))



from typing import Any, Dict, Optional, Protocol, Tuple, Union

import numpy as np
import pandas as pd
import torch as th

from ...data.extractors.tabular_extractor import TabularFeatureExtractor
from ...data.handler.data_processor import DataProcessor
from ...data.iterators.base_iterator import BaseIterator, UdaoIterator
from ...utils.interfaces import UdaoInput
from ..utils.moo_utils import get_default_device

InputVariables = Union[Dict[str, np.ndarray], Dict[str, Any]]
InputParameters = Optional[Dict[str, Any]]


class UdaoFunction(Protocol):
    """A function that can be used for a constraint or an objective."""

    def __call__(
        self,
        input_variables: InputVariables,
        input_parameters: InputParameters = None,
    ) -> th.Tensor:
        """Call the function using the input variables and parameters

        Parameters
        ----------
        input_variables : InputVariables
            a dictionary of the variables inputs
            (one or several values for several variables samples)
        input_parameters : InputParameters, optional
            a dictionary of the non-decision inputs
            (one value per parameter as the value is fixed),
            by default None

        Returns
        -------
        th.Tensor
            the output of the function
        """
        ...


class ModelComponent:
    """A wrapper class for a model paired with
    a DataProcessor, that complies with the UdaoFunction protocol."""

    def __init__(self, data_processor: DataProcessor, model: th.nn.Module) -> None:
        self.data_processor = data_processor
        self.model = model

    def process_data(
        self,
        input_variables: InputVariables,
        input_parameters: InputParameters = None,
    ) -> Tuple[Any, BaseIterator]:
        """Derive the batch input from the input dict."""
        return derive_processed_input(
            self.data_processor,
            input_parameters=input_parameters,
            input_variables=input_variables,
        )

    def __call__(
        self,
        input_variables: InputVariables,
        input_parameters: InputParameters = None,
    ) -> th.Tensor:
        input_data, _ = self.process_data(
            input_parameters=input_parameters, input_variables=input_variables
        )
        return self.model(input_data)

    def to(self, device: th.device) -> None:
        self.model.to(device)


class InaccurateModel(th.nn.Module):
    """A helper class to define an inaccurate model
    as a function of a model and a standard deviation function.

    Parameters
    ----------
    model : th.nn.Module
        The model to be used
    std_func : th.nn.Module
        The standard deviation function associated to the model
    alpha : float
        The factor to multiply the standard deviation
        by before adding it to the model output
    """

    def __init__(
        self, model: th.nn.Module, std_func: th.nn.Module, alpha: float
    ) -> None:
        """ """
        self.model = model
        self.std_func = std_func
        self.alpha = alpha

    def forward(self, x: th.Tensor) -> th.Tensor:
        std = self.std_func(x)
        return self.model(x) + self.alpha * std


def derive_unprocessed_input(
    input_variables: InputVariables,
    input_parameters: InputParameters = None,
    device: Optional[th.device] = None,
) -> Tuple[Dict[str, th.Tensor], Dict[str, Any]]:
    """Derive the input data from the input values,
    in the case where the input is not processed by a DataProcessor

    Parameters
    ----------
    input_variables : InputVariables
        a dictionary of the variables inputs
        (one or several values for several variables samples)
    input_parameters : InputParameters, optional
        a dictionary of the non-decision inputs
        (one value per parameter as the value is fixed),
        by default None
    device : th.device, optional
        device on which to put the resulting tensors, by default th.device("cpu")

    Returns
    -------
    Tuple[Dict[str, th.Tensor], Dict[str, Any]]
        the input data for the model,
        with the variables inputs as tensors and the non-decision inputs as values
    """
    device = device or get_default_device()

    variable_sample = input_variables[list(input_variables.keys())[0]]
    if isinstance(variable_sample, np.ndarray):
        n_items = len(variable_sample)
    else:
        n_items = 1
        input_variables = {k: np.array([v]) for k, v in input_variables.items()}
    input_parameters_values = {
        k: th.tensor([v] * n_items).to(device)
        for k, v in (input_parameters or {}).items()
    }
    return {
        name: th.tensor(value, device=device).float()
        for name, value in input_variables.items()
    }, input_parameters_values


def derive_processed_input(
    data_processor: DataProcessor[UdaoIterator],
    input_variables: InputVariables,
    input_parameters: InputParameters = None,
    device: Optional[th.device] = None,
) -> Tuple[UdaoInput, UdaoIterator]:
    """Derive the batch input from the input dict using a DataProcessor

    Parameters
    ----------
    input_non_decision : Dict[str, Any]
        The fixed values for the non-decision inputs

    input_variables : Dict[str, np.ndarray] | Dict[str, Any]
        The values for the variables inputs:
        - a batch of all values if the variable is a numpy array
        - a batch of a single value if the variable is a single value

    Returns
    -------
    Tuple[UdaoInput, BaseIterator]
        - The batch input for the model
        - The iterator used to generate the batch input
    """
    device = device or get_default_device()
    variable_sample = input_variables[list(input_variables.keys())[0]]
    if isinstance(variable_sample, np.ndarray):
        n_items = len(input_variables[list(input_variables.keys())[0]])
    else:
        n_items = 1
        input_variables = {k: [v] for k, v in input_variables.items()}
    keys = [f"{i}" for i in range(n_items)]
    objective_values = {}
    # hack to get dummy objective values when using a tabular extractor
    if (obj_extractor := data_processor.feature_extractors["objectives"]) is not None:
        if isinstance(obj_extractor, TabularFeatureExtractor) and obj_extractor.columns:
            objective_values = {k: 0 for k in obj_extractor.columns}

    pd_input = pd.DataFrame.from_dict(
        {
            **{k: [v] * n_items for k, v in (input_parameters or {}).items()},
            **objective_values,
            **input_variables,
            "id": keys,
        }
    )
    pd_input.set_index("id", inplace=True, drop=False)
    iterator = data_processor.make_iterator(pd_input, keys, split="test")
    dataloader = iterator.get_dataloader(batch_size=n_items)
    batch_input, _ = next(iter(dataloader))
    return batch_input.to(device) if device else batch_input, iterator



import numpy as np
import pandas as pd
import pytest

from ...containers.tabular_container import TabularContainer
from ...preprocessors.one_hot_preprocessor import OneHotPreprocessor


@pytest.fixture
def sample_container() -> TabularContainer:
    return TabularContainer(
        pd.DataFrame(
            {
                "color": ["red", "green", "blue", "green"],
                "size": ["S", "M", "L", "S"],
                "price": [10, 15, 20, 15],
            }
        )
    )


@pytest.fixture
def one_hot_encoder_preprocessor() -> OneHotPreprocessor:
    return OneHotPreprocessor(["color", "size"])


class TestOneHotPreprocessor:
    def test_preprocess(
        self,
        one_hot_encoder_preprocessor: OneHotPreprocessor,
        sample_container: TabularContainer,
    ) -> None:
        # Act
        container_encoded = one_hot_encoder_preprocessor.preprocess(
            sample_container, split="train"
        )
        df_encoded = container_encoded.data

        # Assert
        assert "color" not in df_encoded.columns
        assert "size" not in df_encoded.columns
        assert one_hot_encoder_preprocessor.encoded_feature_names is not None
        assert len(one_hot_encoder_preprocessor.encoded_feature_names) > 0
        for feature in one_hot_encoder_preprocessor.encoded_feature_names:
            assert feature in df_encoded.columns

    def test_inverse_preprocess(
        self,
        one_hot_encoder_preprocessor: OneHotPreprocessor,
        sample_container: TabularContainer,
    ) -> None:
        container_encoded = one_hot_encoder_preprocessor.preprocess(
            sample_container, split="train"
        )

        container_original = one_hot_encoder_preprocessor.inverse_preprocess(
            container_encoded
        )
        df_original = container_original.data

        np.testing.assert_array_equal(
            df_original.sort_index(axis=1).values,
            sample_container.data.sort_index(axis=1).values,
        )

    def test_inverse_preprocess_without_preprocess(
        self,
        one_hot_encoder_preprocessor: OneHotPreprocessor,
        sample_container: TabularContainer,
    ) -> None:
        # Assert
        with pytest.raises(ValueError):
            one_hot_encoder_preprocessor.inverse_preprocess(sample_container)



from typing import List, Optional, cast

import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from ..utils.utils import DatasetType
from .base_preprocessor import T, TrainedPreprocessor


class OneHotPreprocessor(TrainedPreprocessor):
    """One-hot encode the specified categorical features.

    Parameters
    ----------
    encoder : OneHotEncoder
        An instance of sklearn's OneHotEncoder
    categorical_features : List[str]
        The list of column names to encode.
    """

    def __init__(
        self,
        categorical_features: Optional[List[str]],
        data_key: str = "data",
    ) -> None:
        self.encoder = OneHotEncoder(
            sparse_output=False,
            handle_unknown="ignore",
        )
        self.categorical_features = categorical_features
        self.df_key = data_key
        self.encoded_feature_names: Optional[np.ndarray] = None

    def preprocess(self, container: T, split: DatasetType) -> T:
        """Apply one-hot encoding to the data in the container.

        Parameters
        ----------
        container : T
            Child of BaseContainer
        split : DatasetType
            Train or other (val, test).

        Returns
        -------
        T
            Child of BaseContainer with one-hot encoded data.
        """
        container = container.copy()
        df: pd.DataFrame = container.__getattribute__(self.df_key)

        if split == "train":
            if self.categorical_features is None:
                self.categorical_features = df.select_dtypes(
                    include=["category", "object"]
                ).columns.tolist()
            self.encoder.fit(df[self.categorical_features].to_numpy())
            self.encoded_feature_names = self.encoder.get_feature_names_out(
                self.categorical_features
            )

        if self.categorical_features is None:
            raise ValueError(
                "Categorical features not found. "
                "Please preprocess the data before preprocessing."
            )
        transformed_data = cast(
            np.ndarray, self.encoder.transform(df[self.categorical_features])
        )

        encoded_df = pd.DataFrame(
            data=transformed_data, index=df.index, columns=self.encoded_feature_names
        )

        container.__setattr__(
            self.df_key,
            pd.concat([df.drop(self.categorical_features, axis=1), encoded_df], axis=1),
        )
        return container

    def inverse_preprocess(self, container: T) -> T:
        """Reverse the one-hot encoding process on the container's data.

        Parameters
        ----------
        container : T
            Child of BaseContainer with one-hot encoded data.

        Returns
        -------
        T
            Child of BaseContainer with the original categorical data.
        """
        container = container.copy()
        df: pd.DataFrame = container.__getattribute__(self.df_key)
        if self.encoded_feature_names is None:
            raise ValueError(
                "Encoded feature names not found. "
                "Please preprocess the data before inverse preprocessing."
            )
        encoded_columns = [
            col
            for col in df.columns
            if col.startswith(tuple(self.encoded_feature_names))
        ]
        encoded_data = df[encoded_columns]

        original_data = self.encoder.inverse_transform(encoded_data)
        original_df = pd.DataFrame(
            original_data,
            index=df.index,
            columns=self.categorical_features,
            dtype=pd.CategoricalDtype,
        )

        container.__setattr__(
            self.df_key,
            pd.concat([df.drop(encoded_columns, axis=1), original_df], axis=1),
        )
        return container



from typing import Sequence, Tuple

import torch as th

from ....data.containers.tabular_container import TabularContainer
from ....data.iterators.base_iterator import UdaoIterator
from ....utils.interfaces import (
    UdaoEmbedInput,
    UdaoEmbedItemShape,
    UdaoInput,
    UdaoItemShape,
)


class DummyUdaoIterator(UdaoIterator[UdaoInput, UdaoItemShape]):
    def __init__(
        self,
        keys: Sequence[str],
        tabular_features: TabularContainer,
        objectives: TabularContainer,
    ) -> None:
        super().__init__(keys, tabular_features=tabular_features, objectives=objectives)

    def _getitem(self, idx: int) -> Tuple[UdaoInput, th.Tensor]:
        key = self.keys[idx]
        return (
            UdaoInput(
                th.tensor(self.tabular_features.get(key), dtype=self.tensors_dtype)
            ),
            th.tensor(self.objectives.get(key), dtype=self.tensors_dtype),
        )

    @property
    def shape(self) -> UdaoItemShape:
        return UdaoItemShape(
            feature_names=list(self.tabular_features.data.columns),
            output_names=list(self.objectives.data.columns),
        )

    @staticmethod
    def collate(
        items: Sequence[Tuple[UdaoInput, th.Tensor]]
    ) -> Tuple[UdaoInput, th.Tensor]:
        features = UdaoInput(th.vstack([item[0].features for item in items]))
        objectives = th.vstack([item[1] for item in items])
        return features, objectives


class DummyUdaoEmbedIterator(UdaoIterator[UdaoEmbedInput, UdaoEmbedItemShape]):
    def __init__(
        self,
        keys: Sequence[str],
        tabular_features: TabularContainer,
        objectives: TabularContainer,
        embedding_features: TabularContainer,
    ) -> None:
        self.embedding_features = embedding_features
        super().__init__(keys, tabular_features=tabular_features, objectives=objectives)

    def _getitem(self, idx: int) -> Tuple[UdaoEmbedInput, th.Tensor]:
        key = self.keys[idx]
        return (
            UdaoEmbedInput(
                embedding_input=th.tensor(
                    self.embedding_features.get(key), dtype=self.tensors_dtype
                ),
                features=th.tensor(
                    self.tabular_features.get(key), dtype=self.tensors_dtype
                ),
            ),
            th.tensor(self.objectives.get(key), dtype=self.tensors_dtype),
        )

    @property
    def shape(self) -> UdaoEmbedItemShape:
        return UdaoEmbedItemShape(
            embedding_input_shape=self.embedding_features.data.shape[1],
            feature_names=list(self.tabular_features.data.columns),
            output_names=list(self.objectives.data.columns),
        )

    @staticmethod
    def collate(
        items: Sequence[Tuple[UdaoEmbedInput, th.Tensor]]
    ) -> Tuple[UdaoEmbedInput, th.Tensor]:
        embedding_input = th.vstack([item[0].embedding_input for item in items])
        features = th.vstack([item[0].features for item in items])
        objectives = th.vstack([item[1] for item in items])
        return (
            UdaoEmbedInput(embedding_input=embedding_input, features=features),
            objectives,
        )



from abc import ABC
from dataclasses import dataclass
from typing import Any, Dict, Literal, Optional, Sequence

import dgl
import torch as th
import torch.nn as nn

from ...utils.interfaces import UdaoEmbedItemShape
from ...utils.logging import logger
from .base_embedder import BaseEmbedder
from .layers.iso_bn import IsoBN

NormalizerType = Literal["BN", "LN", "IsoBN"]


class BaseGraphEmbedder(BaseEmbedder, ABC):
    """Base class for Embedder networks.
    Takes care of preparing the input features for the
    embedding layer, and normalizing the output embedding.

    Parameters
    ----------
    net_params : EmbedderParams
    """

    @dataclass
    class Params(BaseEmbedder.Params):
        input_size: int  # depends on the data
        """The size of the input features, except for the type of operation.
        If type is provided, the input size is increased at init
        by the type embedding dimension.
        """
        n_op_types: Optional[int]  # depends on the data
        """The number of operation types."""
        op_groups: Sequence[str]
        """The groups of operation features to be included in the embedding."""
        type_embedding_dim: Optional[int]
        """The dimension of the operation type embedding."""
        embedding_normalizer: Optional[NormalizerType]
        """Name of the normalizer to use for the output embedding."""

    @classmethod
    def from_iterator_shape(
        cls,
        iterator_shape: UdaoEmbedItemShape[Dict[str, int]],
        **kwargs: Any,
    ) -> "BaseGraphEmbedder":
        embedding_input_shapes = iterator_shape.embedding_input_shape
        op_groups = [name for name in embedding_input_shapes.keys()]
        input_size = sum(
            [embedding_input_shapes[name] for name in op_groups if name != "type"]
        )
        n_op_types = None
        if "type" in op_groups:
            n_op_types = iterator_shape.embedding_input_shape["type"]
        params_dict = {
            "op_groups": op_groups,
            "n_op_types": n_op_types,
            "input_size": input_size,
            **kwargs,
        }
        if any((name not in cls.Params.__dataclass_fields__) for name in params_dict):
            for name in params_dict:
                if name not in cls.Params.__dataclass_fields__:
                    logger.debug(f"{name} is not a valid parameter for {cls.__name__}")
            raise ValueError(f"Some parameters are not valid for {cls.__name__} Params")
        return cls(cls.Params(**params_dict))

    def __init__(self, net_params: Params) -> None:
        super().__init__(net_params)
        self.input_size = net_params.input_size

        op_groups = net_params.op_groups
        self.op_type = "type" in op_groups
        self.op_cbo = "cbo" in op_groups
        self.op_enc = "op_enc" in op_groups
        if self.op_type:
            if net_params.n_op_types is None or net_params.type_embedding_dim is None:
                raise ValueError(
                    "n_op_types and type_embedding_dim must be provided "
                    "if `type` is included in op_groups"
                )
            self.op_embedder = nn.Embedding(
                net_params.n_op_types, net_params.type_embedding_dim
            )
            self.input_size += net_params.type_embedding_dim
        self.out_norm: Optional[nn.Module] = None
        if net_params.embedding_normalizer is None:
            self.out_norm = None
        elif net_params.embedding_normalizer == "BN":
            self.out_norm = nn.BatchNorm1d(self.embedding_size)
        elif net_params.embedding_normalizer == "LN":
            self.out_norm = nn.LayerNorm(self.embedding_size)
        elif net_params.embedding_normalizer == "IsoBN":
            self.out_norm = IsoBN(self.embedding_size)
        else:
            raise ValueError(net_params.embedding_normalizer)

    def concatenate_op_features(self, g: dgl.DGLGraph) -> th.Tensor:
        """Concatenate the operation features into a single tensor.

        Parameters
        ----------
        g : dgl.DGLGraph
            Input graph

        Returns
        -------
        th.Tensor
            output tensor of shape (num_nodes, input_size)
        """
        op_list = []
        if self.op_type:
            op_list.append(self.op_embedder(g.ndata["op_gid"]))
        if self.op_cbo:
            op_list.append(g.ndata["cbo"])
        if self.op_enc:
            op_list.append(g.ndata["op_enc"])
        op_tensor = th.cat(op_list, dim=1) if len(op_list) > 1 else op_list[0]
        return op_tensor

    def normalize_embedding(self, embedding: th.Tensor) -> th.Tensor:
        """Normalizes the embedding."""
        if self.out_norm is not None:
            embedding = self.out_norm(embedding)
        return embedding



from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

import numpy as np
from gensim.corpora import Dictionary
from gensim.models import TfidfModel, Word2Vec
from gensim.models.phrases import Phraser, Phrases

from .base_predicate_embedder import BasePredicateEmbedder


@dataclass
class Word2VecParams:
    """
    Parameters to pass to the gensim.Word2Vec model
    (ref: https://radimrehurek.com/gensim/models/word2vec.html)
    """

    min_count: int = 1
    window: int = 3
    vec_size: int = 32
    alpha: float = 0.025
    sample: float = 0.1
    min_alpha: float = 0.0007
    workers: int = 1  # mp.cpu_count() - 1
    seed: int = 42
    epochs: int = 10


class Word2VecEmbedder(BasePredicateEmbedder):
    """
    A class to embed query plans using Word2Vec.
    The embedding is computed as the average of the word
    embeddings of the words in the query plan.

    To use it:
    - first call fit_transform on a list of training query plans,
    - then call transform on a list of query plans.

    N.B. To ensure reproducibility, several things need to be done:
    - set the seed in the Word2VecParams
    - set the PYTHONHASHSEED
    - set the number of workers to 1

    Parameters
    ----------
    w2v_params : Word2VecParams
        Parameters to pass to the gensim.Word2Vec model
        (ref: https://radimrehurek.com/gensim/models/word2vec.html)
    """

    def __init__(self, w2v_params: Optional[Word2VecParams] = None) -> None:
        if w2v_params is None:
            w2v_params = Word2VecParams()
        self.w2v_params = w2v_params
        self.w2v_model = Word2Vec(
            min_count=w2v_params.min_count,
            window=w2v_params.window,  # 3
            vector_size=w2v_params.vec_size,  # 32
            sample=w2v_params.sample,
            alpha=w2v_params.alpha,  # 0.025
            min_alpha=w2v_params.min_alpha,
            workers=w2v_params.workers,
            seed=w2v_params.seed,
            epochs=w2v_params.epochs,
        )
        self._bigram_model: Optional[Phraser] = None
        self.dictionary = Dictionary()
        self.tfidf_model = TfidfModel(smartirs="ntc")

    def _get_encodings_from_corpus(
        self, bow_corpus: List[List[Tuple[int, int]]]
    ) -> np.ndarray:
        """Compute the encodings from a bow_corpus"""
        return np.array(
            [
                np.mean(
                    [
                        self.w2v_model.wv.get_vector(self.dictionary[wid], norm=True)
                        * freq
                        for wid, freq in wt
                    ],
                    0,
                )
                for wt in self.tfidf_model[bow_corpus]
            ]
        )

    def fit_transform(
        self, training_texts: Sequence[str], epochs: Optional[int] = None
    ) -> np.ndarray:
        """Train the Word2Vec model on the training texts and return the embeddings.

        Parameters
        ----------
        training_texts : Sequence[str]
            list of training texts
        epochs : int, optional
            number of epochs for training the model, by default will use the value
            in the Word2VecParams.

        Returns
        -------
        np.ndarray
            Embeddings of the training plans
        """
        if epochs is None:
            epochs = self.w2v_params.epochs
        training_sentences = [row.split() for row in training_texts]
        phrases = Phrases(training_sentences, min_count=30)  # Extract parameter
        self._bigram_model = Phraser(phrases)
        training_descriptions = self._bigram_model[training_sentences]
        self.w2v_model.build_vocab(training_descriptions)
        self.w2v_model.train(
            training_descriptions,
            total_examples=self.w2v_model.corpus_count,
            epochs=epochs,
            report_delay=1,
        )
        # print(f"get {self.w2v_model.wv.vectors.shape[0]} words from word2vec")
        bow_corpus: List[List[Tuple[int, int]]] = [
            self.dictionary.doc2bow(doc, allow_update=True, return_missing=False)  # type: ignore
            for doc in training_descriptions
        ]
        self.tfidf_model.initialize(bow_corpus)
        # print(f"get {len(self.tfidf.idfs)} words from tfidf")
        return self._get_encodings_from_corpus(bow_corpus=bow_corpus)

    def transform(self, texts: Sequence[str]) -> np.ndarray:
        """Transform a list of query plans into embeddings.

        Parameters
        ----------
        texts : Sequence[str]
            list of texts to transform
        epochs : int, optional
            number of epochs for infering a document's embedding,
            by default will use the value in the Word2VecParams.

        Returns
        -------
        np.ndarray
            Embeddings of the query plans

        Raises
        ------
        ValueError
            If the model has not been trained
        """
        sentences = [row.split() for row in texts]
        if self._bigram_model is None:
            raise ValueError("Must call fit_transform before calling transform")
        descriptions = self._bigram_model[sentences]

        bow_corpus: List[List[Tuple[int, int]]] = [
            self.dictionary.doc2bow(doc, allow_update=False, return_missing=False)  # type: ignore
            for doc in descriptions
        ]

        return self._get_encodings_from_corpus(bow_corpus=bow_corpus)



from dataclasses import dataclass, field
from typing import Optional, Union

import numpy as np


@dataclass
class Variable:
    """Variable to optimize."""

    pass


@dataclass
class NumericVariable(Variable):
    lower: Union[int, float]
    upper: Union[int, float]

    def __post_init__(self) -> None:
        if self.lower > self.upper:
            raise ValueError(
                f"ERROR: the lower bound of variable {self}"
                " is greater than its upper bound!"
            )


@dataclass
class IntegerVariable(NumericVariable):
    """Numeric variable with integer values."""

    lower: int
    upper: int


@dataclass
class FloatVariable(NumericVariable):
    """Numeric variable with float values."""

    lower: float
    upper: float


@dataclass
class BoolVariable(IntegerVariable):
    """Boolean variable."""

    lower: int = field(default=0, init=False)
    upper: int = field(default=1, init=False)


@dataclass
class EnumVariable(Variable):
    """Categorical variable (non-numeric)"""

    values: list


def get_random_variable_values(
    var: Variable, n_samples: int, seed: Optional[int] = None
) -> np.ndarray:
    if seed is not None:
        np.random.seed(seed)
    if isinstance(var, FloatVariable):
        return np.random.uniform(var.lower, var.upper, n_samples)
    elif isinstance(var, IntegerVariable):
        return np.random.randint(int(var.lower), int(var.upper + 1), size=n_samples)
    elif isinstance(var, EnumVariable):
        inds = np.random.randint(0, len(var.values), size=n_samples)
        return np.array(var.values)[inds]
    else:
        raise NotImplementedError(f"ERROR: variable type {type(var)} is not supported!")
